# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import urllib.request
from enum import Enum

import aif360
import aif360.datasets
import numpy as np
import pandas as pd
from aif360.algorithms.preprocessing.optim_preproc_helpers.data_preproc_functions import (
    load_preproc_data_compas,
)

import lale.datasets
import lale.datasets.openml
import lale.lib.aif360.util

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


def fetch_adult_df(preprocess=False):
    """
    Fetch the `adult`_ dataset from OpenML and add `fairness_info`.
    It contains information about individuals from the 1994 U.S. census.
    The prediction task is a binary classification on whether the
    income of a person exceeds 50K a year. Without preprocessing,
    the dataset has 48,842 rows and 14 columns. There are two
    protected attributes, sex and race, and the disparate impact is
    0.23. The data includes both categorical and numeric columns, and
    has some missing values.

    .. _`adult`: https://www.openml.org/d/179

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      impute missing values;
      encode protected attributes in X as 0 or 1 to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate favorable outcomes;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "adult", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    if preprocess:
        sex = pd.Series(orig_X["sex_Male"] == 1, dtype=np.float64)
        race = pd.Series(orig_X["race_White"] == 1, dtype=np.float64)
        dropped_X = orig_X.drop(
            labels=[
                "race_Amer-Indian-Eskimo",
                "race_Asian-Pac-Islander",
                "race_Black",
                "race_Other",
                "race_White",
                "sex_Female",
                "sex_Male",
            ],
            axis=1,
        )
        encoded_X = dropped_X.assign(sex=sex, race=race)
        assert not encoded_X.isna().any().any()
        assert not orig_y.isna().any().any()
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "sex", "reference_group": [1]},
                {"feature": "race", "reference_group": [1]},
            ],
        }
        return encoded_X, orig_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": [">50K"],
            "protected_attributes": [
                {"feature": "race", "reference_group": ["White"]},
                {"feature": "sex", "reference_group": ["Male"]},
            ],
        }
        return orig_X, orig_y, fairness_info


def fetch_bank_df(preprocess=False):
    """
    Fetch the `bank-marketing`_ dataset from OpenML and add `fairness_info`.

    It contains information from marketing campaigns of a Portuguise
    bank.  The prediction task is a binary classification on whether
    the client will subscribe a term deposit. Without preprocessing,
    the dataset has 45,211 rows and 16 columns. There is one protected
    attribute, age, and the disparate impact of 0.84. The data
    includes both categorical and numeric columns, with no missing
    values.

    .. _`bank-marketing`: https://www.openml.org/d/1461

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate favorable outcomes;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "bank-marketing", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index().astype(np.float64)
    column_map = {
        "v1": "age",
        "v2": "job",
        "v3": "marital",
        "v4": "education",
        "v5": "default",
        "v6": "balance",
        "v7": "housing",
        "v8": "loan",
        "v9": "contact",
        "v10": "day",
        "v11": "month",
        "v12": "duration",
        "v13": "campaign",
        "v14": "pdays",
        "v15": "previous",
        "v16": "poutcome",
    }
    if preprocess:

        def map_col(col):
            if col.find("_") == -1:
                return column_map[col]
            prefix, suffix = col.split("_")
            return column_map[prefix] + "_" + suffix

        orig_X.columns = [map_col(col) for col in orig_X.columns]
        age = pd.Series(orig_X["age"] >= 25, dtype=np.float64)
        encoded_X = orig_X.assign(age=age)
        encoded_y = pd.Series(orig_y == 0, dtype=np.float64, name=orig_y.name)
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "age", "reference_group": [1]},
            ],
        }
        return encoded_X, encoded_y, fairness_info
    else:
        orig_X.columns = [column_map[col] for col in orig_X.columns]
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "age", "reference_group": [[25, 1000]]},
            ],
        }
        return orig_X, orig_y, fairness_info


def fetch_tae_df(preprocess=False):
    """
    Fetch the `tae`_ dataset from OpenML and add `fairness_info`.

    It contains information from teaching assistant (TA) evaluations.
    at the University of Wisconsin--Madison.
    The prediction task is a classification on the type
    of rating a TA receives (1=Low, 2=Medium, 3=High). Without preprocessing,
    the dataset has 151 rows and 5 columns. There is one protected
    attributes, "whether_of_not_the_ta_is_a_native_english_speaker" [sic],
    and the disparate impact of 0.45. The data
    includes both categorical and numeric columns, with no missing
    values.

    .. _`tae`: https://www.openml.org/d/48

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged group
      ("native_english_speaker");
      encode labels in y as 0 or 1 to indicate favorable outcomes;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protecteded attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "tae", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index().astype(np.float64)
    orig_y = pd.concat([train_y, test_y]).sort_index().astype(np.float64)

    if preprocess:
        native_english_speaker = pd.Series(
            orig_X["whether_of_not_the_ta_is_a_native_english_speaker_1"] == 1,
            dtype=np.float64,
        )
        dropped_X = orig_X.drop(
            labels=[
                "whether_of_not_the_ta_is_a_native_english_speaker_1",
                "whether_of_not_the_ta_is_a_native_english_speaker_2",
            ],
            axis=1,
        )
        encoded_X = dropped_X.assign(native_english_speaker=native_english_speaker)
        encoded_y = pd.Series(orig_y == 2, dtype=np.float64)
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "native_english_speaker", "reference_group": [1]},
            ],
        }
        return encoded_X, encoded_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": [3],
            "protected_attributes": [
                {
                    "feature": "whether_of_not_the_ta_is_a_native_english_speaker",
                    "reference_group": [1],
                },
            ],
        }
        return orig_X, orig_y, fairness_info


# COMPAS HELPERS
def _get_compas_filename(violent_recidivism=False):
    violent_tag = ""
    if violent_recidivism:
        violent_tag = "-violent"
    filename = f"compas-scores-two-years{violent_tag}.csv"
    return filename


def _get_compas_filepath(filename):
    directory = os.path.join(
        os.path.dirname(os.path.abspath(aif360.__file__)), "data", "raw", "compas"
    )
    return os.path.join(
        directory,
        filename,
    )


def _try_download_compas(violent_recidivism=False):
    filename = _get_compas_filename(violent_recidivism=violent_recidivism)
    filepath = _get_compas_filepath(filename)
    csv_exists = os.path.exists(filepath)
    if not csv_exists:
        urllib.request.urlretrieve(
            f"https://raw.githubusercontent.com/propublica/compas-analysis/master/{filename}",
            filepath,
        )


def _get_pandas_and_fairness_info_from_compas_dataset(dataset):
    X, y = lale.lib.aif360.util.dataset_to_pandas(dataset)
    assert X is not None

    fairness_info = {
        "favorable_labels": [0],
        "protected_attributes": [
            {"feature": "sex", "reference_group": [1]},
            {"feature": "race", "reference_group": [1]},
        ],
    }
    return X, y, fairness_info


def _get_dataframe_from_compas_csv(violent_recidivism=False):
    filename = _get_compas_filename(violent_recidivism=violent_recidivism)
    filepath = _get_compas_filepath(filename)
    try:
        df = pd.read_csv(filepath, index_col="id", na_values=[])
    except IOError as err:
        # In practice should not get here because of the _try_download_compas call above, but adding failure logic just in case
        logger.error("IOError: {}".format(err))
        logger.error("To use this class, please download the following file:")
        logger.error(
            "\n\thttps://raw.githubusercontent.com/propublica/compas-analysis/master/compas-scores-two-years.csv"
        )
        logger.error("\nand place it, as-is, in the folder:")
        logger.error("\n\t{}\n".format(os.path.abspath(os.path.dirname(filepath))))
        import sys

        sys.exit(1)
    if violent_recidivism:
        # violent recidivism dataset includes extra label column for some reason
        df = pd.DataFrame(
            df,
            columns=list(
                filter(lambda x: x != "two_year_recid.1", df.columns.tolist())
            ),
        ).sort_index()
    return df


def _perform_default_preprocessing(df):
    return df[
        (df.days_b_screening_arrest <= 30)
        & (df.days_b_screening_arrest >= -30)
        & (df.is_recid != -1)
        & (df.c_charge_degree != "O")
        & (df.score_text != "N/A")
    ]


def _perform_custom_preprocessing(df):
    """The custom pre-processing function is adapted from
    https://github.com/fair-preprocessing/nips2017/blob/master/compas/code/Generate_Compas_Data.ipynb
    """

    df = df[
        [
            "age",
            "c_charge_degree",
            "race",
            "age_cat",
            "score_text",
            "sex",
            "priors_count",
            "days_b_screening_arrest",
            "decile_score",
            "is_recid",
            "two_year_recid",
            "c_jail_in",
            "c_jail_out",
        ]
    ]

    # Indices of data samples to keep
    ix = df["days_b_screening_arrest"] <= 30
    ix = (df["days_b_screening_arrest"] >= -30) & ix
    ix = (df["is_recid"] != -1) & ix
    ix = (df["c_charge_degree"] != "O") & ix
    ix = (df["score_text"] != "N/A") & ix
    df = df.loc[ix, :]
    df["length_of_stay"] = (
        pd.to_datetime(df["c_jail_out"]) - pd.to_datetime(df["c_jail_in"])
    ).apply(lambda x: x.days)

    # Restrict races to African-American and Caucasian
    dfcut = df.loc[
        ~df["race"].isin(["Native American", "Hispanic", "Asian", "Other"]), :
    ]

    # Restrict the features to use
    dfcutQ = dfcut[
        [
            "sex",
            "race",
            "age_cat",
            "c_charge_degree",
            "score_text",
            "priors_count",
            "is_recid",
            "two_year_recid",
            "length_of_stay",
        ]
    ].copy()

    # Quantize priors count between 0, 1-3, and >3
    def quantizePrior(x):
        if x <= 0:
            return "0"
        elif 1 <= x <= 3:
            return "1 to 3"
        else:
            return "More than 3"

    # Quantize length of stay
    def quantizeLOS(x):
        if x <= 7:
            return "<week"
        if 8 < x <= 93:
            return "<3months"
        else:
            return ">3 months"

    # Quantize length of stay
    def adjustAge(x):
        if x == "25 - 45":
            return "25 to 45"
        else:
            return x

    # Quantize score_text to MediumHigh
    def quantizeScore(x):
        if (x == "High") | (x == "Medium"):
            return "MediumHigh"
        else:
            return x

    def group_race(x):
        if x == "Caucasian":
            return 1.0
        else:
            return 0.0

    dfcutQ["priors_count"] = dfcutQ["priors_count"].apply(lambda x: quantizePrior(x))
    dfcutQ["length_of_stay"] = dfcutQ["length_of_stay"].apply(lambda x: quantizeLOS(x))
    dfcutQ["score_text"] = dfcutQ["score_text"].apply(lambda x: quantizeScore(x))
    dfcutQ["age_cat"] = dfcutQ["age_cat"].apply(lambda x: adjustAge(x))

    # Recode sex and race
    dfcutQ["sex"] = dfcutQ["sex"].replace({"Female": 1.0, "Male": 0.0})
    dfcutQ["race"] = dfcutQ["race"].apply(lambda x: group_race(x))

    features = [
        "two_year_recid",
        "sex",
        "race",
        "age_cat",
        "priors_count",
        "c_charge_degree",
    ]

    # Pass vallue to df
    df = dfcutQ[features]

    return df


def _get_pandas_and_fairness_info_from_compas_csv(violent_recidivism=False):
    df = _get_dataframe_from_compas_csv(violent_recidivism=violent_recidivism)
    # preprocessing steps performed by ProPublica team, even in the preprocess=False case
    df = _perform_default_preprocessing(df)
    X = pd.DataFrame(
        df, columns=list(filter(lambda x: x != "two_year_recid", df.columns.tolist()))
    ).sort_index()
    y = pd.Series(
        df["two_year_recid"], name="two_year_recid", dtype=np.float64
    ).sort_index()
    fairness_info = {
        "favorable_labels": [0],
        "protected_attributes": [
            {"feature": "sex", "reference_group": ["Female"]},
            {"feature": "race", "reference_group": ["Caucasian"]},
        ],
    }
    return X, y, fairness_info


def fetch_compas_df(preprocess=False):
    """
    Fetch the `compas-two-years`_ dataset, also known as ProPublica recidivism, from GitHub and add `fairness_info`.

    It contains information about individuals with a binary
    classification for recidivism, indicating whether they were
    re-arrested within two years after the first arrest. Without
    preprocessing, the dataset has 6,172 rows and 51 columns.  There
    are two protected attributes, sex and race, and the disparate
    impact is 0.75.  The data includes numeric and categorical columns, with some
    missing values.

    .. _`compas-two-years`: https://github.com/propublica/compas-analysis

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups
      (1 if Female or Caucasian for the corresponding sex and race columns respectively);
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    violent_recidivism = False
    _try_download_compas(violent_recidivism=violent_recidivism)
    if preprocess:
        # Odd finding here: "Female" is a privileged class in the dataset, but the original
        # COMPAS algorithm actually predicted worse outcomes for that class after controlling
        # for other factors. Leaving it as "Female" for now (AIF360 does this by default as well)
        # but potentially worthy of revisiting.
        # See https://www.propublica.org/article/how-we-analyzed-the-compas-recidivism-algorithm
        # and https://github.com/propublica/compas-analysis/blob/master/Compas%20Analysis.ipynb
        # (hunch is that COMPAS was trained on more biased data that is not reproduced in ProPublica's dataset)
        dataset = load_preproc_data_compas()
        # above preprocessing results in a WARNING of "Missing Data: 5 rows removed from CompasDataset."
        # unclear how to resolve at the moment
        return _get_pandas_and_fairness_info_from_compas_dataset(dataset)
    else:
        return _get_pandas_and_fairness_info_from_compas_csv(
            violent_recidivism=violent_recidivism
        )


def fetch_compas_violent_df(preprocess=False):
    """
    Fetch the `compas-two-years-violent`_ dataset, also known as ProPublica violent recidivism, from GitHub and add `fairness_info`.

    It contains information about individuals with a binary
    classification for violent recidivism, indicating whether they were
    re-arrested within two years after the first arrest. Without
    preprocessing, the dataset has 4,020 rows and 51 columns.  There
    are three protected attributes, sex, race, and age, and the disparate
    impact is 0.85.  The data includes numeric and categorical columns, with some
    missing values.

    .. _`compas-two-years-violent`: https://github.com/propublica/compas-analysis

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups
      (1 if Female, Caucasian, or at least 25 for the corresponding sex, race, and
      age columns respectively);
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    violent_recidivism = True
    _try_download_compas(violent_recidivism=violent_recidivism)
    if preprocess:
        # Odd finding here: "Female" is a privileged class in the dataset, but the original
        # COMPAS algorithm actually predicted worse outcomes for that class after controlling
        # for other factors. Leaving it as "Female" for now (AIF360 does this by default as well)
        # but potentially worthy of revisiting.
        # See https://www.propublica.org/article/how-we-analyzed-the-compas-recidivism-algorithm
        # and https://github.com/propublica/compas-analysis/blob/master/Compas%20Analysis.ipynb
        # (hunch is that COMPAS was trained on more biased data that is not reproduced in ProPublica's dataset)

        # Loading violent recidivism dataset using StandardDataset and default settings found in the CompasDataset
        # class since AIF360 lacks a violent recidivism dataset implementation
        df = _get_dataframe_from_compas_csv(violent_recidivism=violent_recidivism)
        default_mappings = {
            "label_maps": [{1.0: "Did recid.", 0.0: "No recid."}],
            "protected_attribute_maps": [
                {0.0: "Male", 1.0: "Female"},
                {1.0: "Caucasian", 0.0: "Not Caucasian"},
            ],
        }
        dataset = aif360.datasets.StandardDataset(
            df=df,
            label_name="two_year_recid",
            favorable_classes=[0],
            protected_attribute_names=["sex", "race"],
            privileged_classes=[[1.0], [1.0]],
            categorical_features=["age_cat", "priors_count", "c_charge_degree"],
            instance_weights_name=None,
            features_to_keep=[
                "sex",
                "age_cat",
                "race",
                "priors_count",
                "c_charge_degree",
                "two_year_recid",
            ],
            features_to_drop=[],
            na_values=[],
            custom_preprocessing=_perform_custom_preprocessing,
            metadata=default_mappings,
        )
        # above preprocessing results in a WARNING of "Missing Data: 5 rows removed from StandardDataset."
        # unclear how to resolve at the moment
        return _get_pandas_and_fairness_info_from_compas_dataset(dataset)
    else:
        return _get_pandas_and_fairness_info_from_compas_csv(
            violent_recidivism=violent_recidivism
        )


def fetch_creditg_df(preprocess=False):
    """
    Fetch the `credit-g`_ dataset from OpenML and add `fairness_info`.

    It contains information about individuals with a binary
    classification into good or bad credit risks. Without
    preprocessing, the dataset has 1,000 rows and 20 columns. There
    are two protected attributs, personal_status/sex and age, and the
    disparate impact is 0.75.  The data includes both categorical and
    numeric columns, with no missing values.

    .. _`credit-g`: https://www.openml.org/d/31

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate favorable outcomes;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "credit-g", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    if preprocess:
        sex = pd.Series(
            (orig_X["personal_status_male div/sep"] == 1)
            | (orig_X["personal_status_male mar/wid"] == 1)
            | (orig_X["personal_status_male single"] == 1),
            dtype=np.float64,
        )
        age = pd.Series(orig_X["age"] > 25, dtype=np.float64)
        dropped_X = orig_X.drop(
            labels=[
                "personal_status_female div/dep/mar",
                "personal_status_male div/sep",
                "personal_status_male mar/wid",
                "personal_status_male single",
            ],
            axis=1,
        )
        encoded_X = dropped_X.assign(sex=sex, age=age)
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "sex", "reference_group": [1]},
                {"feature": "age", "reference_group": [1]},
            ],
        }
        return encoded_X, orig_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": ["good"],
            "protected_attributes": [
                {
                    "feature": "personal_status",
                    "reference_group": [
                        "male div/sep",
                        "male mar/wid",
                        "male single",
                    ],
                },
                {"feature": "age", "reference_group": [[26, 1000]]},
            ],
        }
        return orig_X, orig_y, fairness_info


def fetch_ricci_df(preprocess=False):
    """
    Fetch the `ricci_vs_destefano`_ dataset from OpenML and add `fairness_info`.

    It contains test scores for 2003 New Haven Fire Department
    promotion exams with a binary classification into promotion or no
    promotion.  Without preprocessing, the dataset has 118 rows and 5
    columns.  There is one protected attribute, race, and the
    disparate impact is 0.50.  The data includes both categorical and
    numeric columns, with no missing values.

    .. _`ricci_vs_destefano`: https://www.openml.org/d/42665

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate favorable outcomes;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "ricci", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    if preprocess:
        race = pd.Series(orig_X["race_W"] == 1, dtype=np.float64)
        dropped_X = orig_X.drop(labels=["race_B", "race_H", "race_W"], axis=1)
        encoded_X = dropped_X.assign(race=race)
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [{"feature": "race", "reference_group": [1]}],
        }
        return encoded_X, orig_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": ["Promotion"],
            "protected_attributes": [{"feature": "race", "reference_group": ["W"]}],
        }
        return orig_X, orig_y, fairness_info


def fetch_speeddating_df(preprocess=False):
    """
    Fetch the `SpeedDating`_ dataset from OpenML and add `fairness_info`.

    It contains data gathered from participants in experimental speed dating events
    from 2002-2004 with a binary classification into match or no
    match.  Without preprocessing, the dataset has 8378 rows and 122
    columns.  There are two protected attributes, whether the other candidate has the same
    race and importance of having the same race, and the disparate impact
    is 0.85.  The data includes both categorical and
    numeric columns, with some missing values.

    .. _`SpeedDating`: https://www.openml.org/d/40536

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate favorable outcomes;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "SpeedDating", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    if preprocess:
        importance_same_race = pd.Series(
            orig_X["importance_same_race"] >= 9, dtype=np.float64
        )
        samerace = pd.Series(orig_X["samerace_1"] == 1, dtype=np.float64)
        # drop samerace-related columns
        columns_to_drop = ["samerace_0", "samerace_1"]

        # drop preprocessed columns

        def preprocessed_column_filter(x: str):
            return x.startswith("d_")

        columns_to_drop.extend(list(filter(preprocessed_column_filter, orig_X.columns)))

        # drop has-null columns
        columns_to_drop.extend(["has_null_0", "has_null_1"])

        # drop decision columns

        def decision_column_filter(x: str):
            return x.startswith("decision")

        columns_to_drop.extend(list(filter(decision_column_filter, orig_X.columns)))

        # drop field columns

        def field_column_filter(x: str):
            return x.startswith("field")

        columns_to_drop.extend(list(filter(field_column_filter, orig_X.columns)))

        # drop wave column
        columns_to_drop.append("wave")
        dropped_X = orig_X.drop(labels=columns_to_drop, axis=1)
        encoded_X = dropped_X.assign(
            samerace=samerace, importance_same_race=importance_same_race
        )
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "samerace", "reference_group": [1]},
                {"feature": "importance_same_race", "reference_group": [1]},
            ],
        }
        return encoded_X, orig_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": ["1"],
            "protected_attributes": [
                {"feature": "samerace", "reference_group": ["1"]},
                {"feature": "importance_same_race", "reference_group": [[9, 1000]]},
            ],
        }
        return orig_X, orig_y, fairness_info


def _fetch_boston_housing_df(preprocess=False):
    """
    Fetch the `Boston housing`_ dataset from sklearn and add `fairness info`.

    It contains data about housing values in the suburbs of Boston with various
    features that can be used to perform regression. Without preprocessing,
    the dataset has 506 rows and 14 columns. There is one protected attribute,
    1000(Bk - 0.63)^2 where Bk is the proportion of Blacks by town, and the disparate
    impact is 0.5. The data includes only numeric columns, with no missing values.

    Hiding dataset from public consumption based on issues described at length `here`_

    .. _`Boston housing`: https://scikit-learn.org/0.20/datasets/index.html#boston-house-prices-dataset
    .. _`here`: https://medium.com/@docintangible/racist-data-destruction-113e3eff54a8

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attribute in X as 0 or 1 to indicate privileged groups.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.boston_housing_df(
        test_size=0.33
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    assert train_X is not None
    black_median = np.median(train_X["B"])
    label_median = np.median(train_y)

    if preprocess:
        # 1000(Bk - 0.63)^2 where Bk is the proportion of Blacks by town
        B = pd.Series(orig_X["B"] > black_median, dtype=np.float64)
        encoded_X = orig_X.assign(B=B)
        fairness_info = {
            "favorable_labels": [[-10000.0, label_median]],
            "protected_attributes": [
                {"feature": "B", "reference_group": [0]},
            ],
        }
        return encoded_X, orig_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": [[-10000.0, label_median]],
            "protected_attributes": [
                # 1000(Bk - 0.63)^2 where Bk is the proportion of Blacks by town
                {"feature": "B", "reference_group": [[0.0, black_median]]},
            ],
        }
        return orig_X, orig_y, fairness_info


def fetch_nursery_df(preprocess=False):
    """
    Fetch the `nursery`_ dataset from OpenML and add `fairness_info`.

    It contains data gathered from applicants to public schools in
    Ljubljana, Slovenia during a competitive time period.
    Without preprocessing, the dataset has
    12960 rows and 8 columns.  There is one protected attribute, parents, and the
    disparate impact is 0.46.  The data has categorical columns (with
    numeric ones if preprocessing is applied), with no missing values.

    .. _`nursery`: https://www.openml.org/d/26

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "nursery", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    if preprocess:
        parents = pd.Series(orig_X["parents_usual"] == 0, dtype=np.float64)
        dropped_X = orig_X.drop(
            labels=[
                "parents_great_pret",
                "parents_pretentious",
                "parents_usual",
            ],
            axis=1,
        )
        encoded_X = dropped_X.assign(parents=parents)
        # orig_y == 3 corresponds to "spec_prior"
        encoded_y = pd.Series((orig_y == 3), dtype=np.float64)
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [{"feature": "parents", "reference_group": [1]}],
        }
        return encoded_X, encoded_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": ["spec_prior"],
            "protected_attributes": [
                {
                    "feature": "parents",
                    "reference_group": ["great_pret", "pretentious"],
                }
            ],
        }
        return orig_X, orig_y, fairness_info


def fetch_titanic_df(preprocess=False):
    """
    Fetch the `Titanic`_ dataset from OpenML and add `fairness_info`.

    It contains data gathered from passengers on the Titanic with a binary classification
    into "survived" or "did not survive".  Without preprocessing, the dataset has
    1309 rows and 13 columns.  There is one protected attribute, sex, and the
    disparate impact is 0.26.  The data includes both categorical and
    numeric columns, with some missing values.

    .. _`Titanic`: https://www.openml.org/d/40945

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attributes in X as 0 or 1 to indicate privileged groups;
      and apply one-hot encoding to any remaining features in X that
      are categorical and not protected attributes.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
        "titanic", "classification", astype="pandas", preprocess=preprocess
    )
    orig_X = pd.concat([train_X, test_X]).sort_index()
    orig_y = pd.concat([train_y, test_y]).sort_index()
    if preprocess:
        sex = pd.Series(orig_X["sex_female"] == 1, dtype=np.float64)
        columns_to_drop = ["sex_female", "sex_male"]

        # drop more columns that turn into gigantic one-hot encodings otherwise, like name and cabin

        def extra_categorical_columns_filter(c: str):
            return (
                c.startswith("name")
                or c.startswith("ticket")
                or c.startswith("cabin")
                or c.startswith("home.dest")
            )

        columns_to_drop.extend(
            list(filter(extra_categorical_columns_filter, orig_X.columns))
        )
        dropped_X = orig_X.drop(labels=columns_to_drop, axis=1)
        encoded_X = dropped_X.assign(sex=sex)
        fairness_info = {
            "favorable_labels": [1],
            "protected_attributes": [
                {"feature": "sex", "reference_group": [1]},
            ],
        }
        return encoded_X, orig_y, fairness_info
    else:
        fairness_info = {
            "favorable_labels": ["1"],
            "protected_attributes": [
                {"feature": "sex", "reference_group": ["female"]},
            ],
        }
        return orig_X, orig_y, fairness_info


# MEPS HELPERS
class FiscalYear(Enum):
    FY2015 = 15
    FY2016 = 16


class Panel(Enum):
    PANEL19 = 19
    PANEL20 = 20
    PANEL21 = 21


def _race(row):
    if (row["HISPANX"] == 2) and (
        row["RACEV2X"] == 1
    ):  # non-Hispanic Whites are marked as WHITE; all others as NON-WHITE
        return "White"
    return "Non-White"


def _get_utilization_columns(fiscal_year):
    return [
        f"OBTOTV{fiscal_year.value}",
        f"OPTOTV{fiscal_year.value}",
        f"ERTOT{fiscal_year.value}",
        f"IPNGTD{fiscal_year.value}",
        f"HHTOTD{fiscal_year.value}",
    ]


def _get_total_utilization(row, fiscal_year):
    cols = _get_utilization_columns(fiscal_year)
    return sum(list(map(lambda x: row[x], cols)))


def _should_drop_column(x, fiscal_year):
    utilization_cols = set(_get_utilization_columns(fiscal_year))
    return x in utilization_cols


def _fetch_meps_raw_df(panel, fiscal_year):
    filename = ""
    if fiscal_year == FiscalYear.FY2015:
        assert panel == Panel.PANEL19 or panel == Panel.PANEL20
        filename = "h181.csv"
    elif fiscal_year == FiscalYear.FY2016:
        assert panel == Panel.PANEL21
        filename = "h192.csv"
    else:
        logger.error(f"Unexpected FiscalYear received: {fiscal_year}")
        raise ValueError(f"Unexpected FiscalYear received: {fiscal_year}")
    filepath = os.path.join(
        os.path.dirname(os.path.abspath(aif360.__file__)),
        "data",
        "raw",
        "meps",
        filename,
    )

    try:
        df = pd.read_csv(filepath, sep=",", na_values=[])
    except IOError as err:
        logger.error("IOError: {}".format(err))
        logger.error("To use this class, please follow the instructions found here:")
        logger.error(
            "\n\t{}\n".format(
                "https://github.com/Trusted-AI/AIF360/tree/master/aif360/data/raw/meps"
            )
        )
        logger.error(
            f"\n to download and convert the data and place the final {filename} file, as-is, in the folder:"
        )
        logger.error("\n\t{}\n".format(os.path.abspath(os.path.dirname(filepath))))
        import sys

        sys.exit(1)

    df["RACEV2X"] = df.apply(lambda row: _race(row), axis=1)
    df = df.rename(columns={"RACEV2X": "RACE"})
    df = df[df["PANEL"] == panel.value]

    df["TOTEXP15"] = df.apply(
        lambda row: _get_total_utilization(row, fiscal_year), axis=1
    )
    lessE = df["TOTEXP15"] < 10.0
    df.loc[lessE, "TOTEXP15"] = 0.0
    moreE = df["TOTEXP15"] >= 10.0
    df.loc[moreE, "TOTEXP15"] = 1.0

    df = df.rename(columns={"TOTEXP15": "UTILIZATION"})
    columns_to_drop = set(
        filter(lambda x: _should_drop_column(x, fiscal_year), df.columns.tolist())
    )
    df = df[sorted(set(df.columns.tolist()) - columns_to_drop, key=df.columns.get_loc)]
    X = pd.DataFrame(
        df, columns=list(filter(lambda x: x != "UTILIZATION", df.columns.tolist()))
    ).sort_index()
    y = pd.Series(df["UTILIZATION"], name="UTILIZATION").sort_index()
    fairness_info = {
        "favorable_labels": [1],
        "protected_attributes": [
            {"feature": "RACE", "reference_group": ["White"]},
        ],
    }

    return X, y, fairness_info


def _get_pandas_and_fairness_info_from_meps_dataset(dataset):
    X, y = lale.lib.aif360.util.dataset_to_pandas(dataset)
    fairness_info = {
        "favorable_labels": [1],
        "protected_attributes": [
            {"feature": "RACE", "reference_group": [1]},
        ],
    }
    return X, y, fairness_info


def fetch_meps_panel19_fy2015_df(preprocess=False):
    """
    Fetch a subset of the `MEPS`_ dataset from aif360 and add fairness info.

    It contains information collected on a nationally representative sample
    of the civilian noninstitutionalized population of the United States,
    specifically reported medical expenditures and civilian demographics.
    This dataframe corresponds to data from panel 19 from the year 2015.
    Without preprocessing, the dataframe contains 16578 rows and 1825 columns.
    (With preprocessing the dataframe contains 15830 rows and 138 columns.)
    There is one protected attribute, race, and the disparate impact is 0.496
    if preprocessing is not applied and 0.490 if preprocessing is applied.
    The data includes numeric and categorical columns, with some missing values.

    Note: in order to use this dataset, be sure to follow the instructions
    found in the `AIF360 documentation`_ and accept the corresponding license agreement.

    .. _`MEPS`:  https://meps.ahrq.gov/mepsweb/data_stats/download_data_files_detail.jsp?cboPufNumber=HC-181
    .. _`AIF360 documentation`: https://github.com/Trusted-AI/AIF360/tree/master/aif360/data/raw/meps

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attribute in X corresponding to race as 0 or 1
      to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate faborable outcomes;
      rename columns that are panel or round-specific;
      drop columns such as ID columns that are not relevant to the task at hand;
      and drop rows where features are unknown.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    if preprocess:
        dataset = aif360.datasets.MEPSDataset19()
        return _get_pandas_and_fairness_info_from_meps_dataset(dataset)
    else:
        return _fetch_meps_raw_df(Panel.PANEL19, FiscalYear.FY2015)


def fetch_meps_panel20_fy2015_df(preprocess=False):
    """
    Fetch a subset of the `MEPS`_ dataset from aif360 and add fairness info.

    It contains information collected on a nationally representative sample
    of the civilian noninstitutionalized population of the United States,
    specifically reported medical expenditures and civilian demographics.
    This dataframe corresponds to data from panel 20 from the year 2015.
    Without preprocessing, the dataframe contains 18849 rows and 1825 columns.
    (With preprocessing the dataframe contains 17570 rows and 138 columns.)
    There is one protected attribute, race, and the disparate impact is 0.493
    if preprocessing is not applied and 0.488 if preprocessing is applied.
    The data includes numeric and categorical columns, with some missing values.

    Note: in order to use this dataset, be sure to follow the instructions
    found in the `AIF360 documentation`_ and accept the corresponding license agreement.

    .. _`MEPS`:  https://meps.ahrq.gov/mepsweb/data_stats/download_data_files_detail.jsp?cboPufNumber=HC-181
    .. _`AIF360 documentation`: https://github.com/Trusted-AI/AIF360/tree/master/aif360/data/raw/meps

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attribute in X corresponding to race as 0 or 1
      to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate faborable outcomes;
      rename columns that are panel or round-specific;
      drop columns such as ID columns that are not relevant to the task at hand;
      and drop rows where features are unknown.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    if preprocess:
        dataset = aif360.datasets.MEPSDataset20()
        return _get_pandas_and_fairness_info_from_meps_dataset(dataset)
    else:
        return _fetch_meps_raw_df(Panel.PANEL20, FiscalYear.FY2015)


def fetch_meps_panel21_fy2016_df(preprocess=False):
    """
    Fetch a subset of the `MEPS`_ dataset from aif360 and add fairness info.

    It contains information collected on a nationally representative sample
    of the civilian noninstitutionalized population of the United States,
    specifically reported medical expenditures and civilian demographics.
    This dataframe corresponds to data from panel 20 from the year 2016.
    Without preprocessing, the dataframe contains 17052 rows and 1936 columns.
    (With preprocessing the dataframe contains 15675 rows and 138 columns.)
    There is one protected attribute, race, and the disparate impact is 0.462
    if preprocessing is not applied and 0.451 if preprocessing is applied.
    The data includes numeric and categorical columns, with some missing values.

    Note: in order to use this dataset, be sure to follow the instructions
    found in the `AIF360 documentation`_ and accept the corresponding license agreement.

    .. _`MEPS`:  https://meps.ahrq.gov/mepsweb/data_stats/download_data_files_detail.jsp?cboPufNumber=HC-181
    .. _`AIF360 documentation`: https://github.com/Trusted-AI/AIF360/tree/master/aif360/data/raw/meps

    Parameters
    ----------
    preprocess : boolean, optional, default False

      If True,
      encode protected attribute in X corresponding to race as 0 or 1
      to indicate privileged groups;
      encode labels in y as 0 or 1 to indicate faborable outcomes;
      rename columns that are panel or round-specific;
      drop columns such as ID columns that are not relevant to the task at hand;
      and drop rows where features are unknown.

    Returns
    -------
    result : tuple

      - item 0: pandas Dataframe

          Features X, including both protected and non-protected attributes.

      - item 1: pandas Series

          Labels y.

      - item 3: fairness_info

          JSON meta-data following the format understood by fairness metrics
          and mitigation operators in `lale.lib.aif360`.
    """
    if preprocess:
        dataset = aif360.datasets.MEPSDataset21()
        return _get_pandas_and_fairness_info_from_meps_dataset(dataset)
    else:
        return _fetch_meps_raw_df(Panel.PANEL21, FiscalYear.FY2016)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from typing import Any, Dict

try:
    import tensorflow as tf

    tensorflow_installed = True
except ImportError:
    tensorflow_installed = False

from lale.lib.aif360 import (
    CalibratedEqOddsPostprocessing,
    DisparateImpactRemover,
    PrejudiceRemover,
    fair_stratified_train_test_split,
)
from lale.lib.aif360.adversarial_debiasing import AdversarialDebiasing
from lale.lib.aif360.datasets import fetch_creditg_df
from lale.lib.sklearn import (
    AdaBoostClassifier,
    BaggingClassifier,
    DecisionTreeClassifier,
    LogisticRegression,
    StackingClassifier,
    VotingClassifier,
)


class TestEnsemblesWithAIF360(unittest.TestCase):
    train_X = None
    train_y = None
    test_X = None
    fairness_info: Dict[str, Any] = {"temp": 0}

    @classmethod
    def setUpClass(cls) -> None:
        X, y, fi = fetch_creditg_df(preprocess=True)
        train_X, test_X, train_y, _ = fair_stratified_train_test_split(X, y, **fi)
        cls.train_X = train_X
        cls.train_y = train_y
        cls.test_X = test_X
        cls.fairness_info = fi

    @classmethod
    def _attempt_fit_predict(cls, model):
        trained = model.fit(cls.train_X, cls.train_y)
        trained.predict(cls.test_X)

    def test_bagging_pre_estimator_mitigation_ensemble(self):
        model = DisparateImpactRemover(**self.fairness_info) >> BaggingClassifier(
            base_estimator=DecisionTreeClassifier()
        )
        self._attempt_fit_predict(model)

    def test_bagging_post_estimator_mitigation_ensemble(self):
        model = CalibratedEqOddsPostprocessing(
            **self.fairness_info,
            estimator=BaggingClassifier(base_estimator=DecisionTreeClassifier())
        )
        self._attempt_fit_predict(model)

    def test_bagging_pre_estimator_mitigation_base(self):
        model = BaggingClassifier(
            base_estimator=DisparateImpactRemover(**self.fairness_info)
            >> DecisionTreeClassifier()
        )
        self._attempt_fit_predict(model)

    def test_bagging_in_estimator_mitigation_base(self):
        model = BaggingClassifier(base_estimator=PrejudiceRemover(**self.fairness_info))
        self._attempt_fit_predict(model)

    def test_bagging_in_estimator_mitigation_base_1(self):
        if tensorflow_installed:
            tf.compat.v1.disable_eager_execution()
            model = BaggingClassifier(
                base_estimator=AdversarialDebiasing(**self.fairness_info),
                n_estimators=2,
            )
            self._attempt_fit_predict(model)

    def test_bagging_post_estimator_mitigation_base(self):
        model = BaggingClassifier(
            base_estimator=CalibratedEqOddsPostprocessing(
                **self.fairness_info, estimator=DecisionTreeClassifier()
            )
        )
        self._attempt_fit_predict(model)

    def test_adaboost_pre_estimator_mitigation_ensemble(self):
        model = DisparateImpactRemover(**self.fairness_info) >> AdaBoostClassifier(
            base_estimator=DecisionTreeClassifier()
        )
        self._attempt_fit_predict(model)

    def test_adaboost_post_estimator_mitigation_ensemble(self):
        model = CalibratedEqOddsPostprocessing(
            **self.fairness_info,
            estimator=AdaBoostClassifier(base_estimator=DecisionTreeClassifier())
        )
        self._attempt_fit_predict(model)

    def test_adaboost_pre_estimator_mitigation_base(self):
        model = AdaBoostClassifier(
            base_estimator=DisparateImpactRemover(**self.fairness_info)
            >> DecisionTreeClassifier()
        )
        self._attempt_fit_predict(model)

    def test_adaboost_in_estimator_mitigation_base(self):
        model = AdaBoostClassifier(
            base_estimator=PrejudiceRemover(**self.fairness_info)
        )
        self._attempt_fit_predict(model)

    def test_adaboost_post_estimator_mitigation_base(self):
        model = AdaBoostClassifier(
            base_estimator=CalibratedEqOddsPostprocessing(
                **self.fairness_info, estimator=DecisionTreeClassifier()
            )
        )
        self._attempt_fit_predict(model)

    def test_voting_pre_estimator_mitigation_ensemble(self):
        model = DisparateImpactRemover(**self.fairness_info) >> VotingClassifier(
            estimators=[("dtc", DecisionTreeClassifier()), ("lr", LogisticRegression())]
        )
        self._attempt_fit_predict(model)

    @unittest.skip("TODO: find out why it does not find predict_proba")
    def test_voting_post_estimator_mitigation_ensemble(self):
        model = CalibratedEqOddsPostprocessing(
            **self.fairness_info,
            estimator=VotingClassifier(
                estimators=[
                    ("dtc", DecisionTreeClassifier()),
                    ("lr", LogisticRegression()),
                ]
            )
        )
        self._attempt_fit_predict(model)

    def test_voting_pre_estimator_mitigation_base(self):
        model = VotingClassifier(
            estimators=[
                (
                    "dir+dtc",
                    DisparateImpactRemover(**self.fairness_info)
                    >> DecisionTreeClassifier(),
                ),
                ("lr", LogisticRegression()),
            ]
        )
        self._attempt_fit_predict(model)

    def test_voting_in_estimator_mitigation_base(self):
        model = VotingClassifier(
            estimators=[
                ("pr", PrejudiceRemover(**self.fairness_info)),
                ("lr", LogisticRegression()),
            ]
        )
        self._attempt_fit_predict(model)

    def test_voting_post_estimator_mitigation_base(self):
        model = VotingClassifier(
            estimators=[
                (
                    "dtc+ceop",
                    CalibratedEqOddsPostprocessing(
                        **self.fairness_info, estimator=DecisionTreeClassifier()
                    ),
                ),
                ("lr", LogisticRegression()),
            ]
        )
        self._attempt_fit_predict(model)

    def test_stacking_pre_estimator_mitigation_ensemble(self):
        model = DisparateImpactRemover(**self.fairness_info) >> StackingClassifier(
            estimators=[("dtc", DecisionTreeClassifier()), ("lr", LogisticRegression())]
        )
        self._attempt_fit_predict(model)

    def test_stacking_post_estimator_mitigation_ensemble(self):
        model = CalibratedEqOddsPostprocessing(
            **self.fairness_info,
            estimator=StackingClassifier(
                estimators=[
                    ("dtc", DecisionTreeClassifier()),
                    ("lr", LogisticRegression()),
                ]
            )
        )
        self._attempt_fit_predict(model)

    def test_stacking_pre_estimator_mitigation_base_only(self):
        model = StackingClassifier(
            estimators=[
                (
                    "dir+dtc",
                    DisparateImpactRemover(**self.fairness_info)
                    >> DecisionTreeClassifier(),
                ),
                ("lr", LogisticRegression()),
            ]
        )
        self._attempt_fit_predict(model)

    def test_stacking_pre_estimator_mitigation_base_and_final(self):
        model = StackingClassifier(
            estimators=[
                (
                    "dir+dtc",
                    DisparateImpactRemover(**self.fairness_info)
                    >> DecisionTreeClassifier(),
                ),
                ("lr", LogisticRegression()),
            ],
            final_estimator=DisparateImpactRemover(**self.fairness_info)
            >> DecisionTreeClassifier(),
            passthrough=True,
        )
        self._attempt_fit_predict(model)

    def test_stacking_pre_estimator_mitigation_final_only(self):
        model = StackingClassifier(
            estimators=[
                ("dtc", DecisionTreeClassifier()),
                ("lr", LogisticRegression()),
            ],
            final_estimator=DisparateImpactRemover(**self.fairness_info)
            >> DecisionTreeClassifier(),
            passthrough=True,
        )
        self._attempt_fit_predict(model)

    def test_stacking_in_estimator_mitigation_base_only(self):
        model = StackingClassifier(
            estimators=[
                ("pr", PrejudiceRemover(**self.fairness_info)),
                ("lr", LogisticRegression()),
            ]
        )
        self._attempt_fit_predict(model)

    def test_stacking_in_estimator_mitigation_base_and_final(self):
        model = StackingClassifier(
            estimators=[
                ("pr", PrejudiceRemover(**self.fairness_info)),
                ("lr", LogisticRegression()),
            ],
            final_estimator=PrejudiceRemover(**self.fairness_info),
            passthrough=True,
        )
        self._attempt_fit_predict(model)

    def test_stacking_in_estimator_mitigation_final_only(self):
        model = StackingClassifier(
            estimators=[
                ("dtc", DecisionTreeClassifier()),
                ("lr", LogisticRegression()),
            ],
            final_estimator=PrejudiceRemover(**self.fairness_info),
            passthrough=True,
        )
        self._attempt_fit_predict(model)

    def test_stacking_post_estimator_mitigation_base_only(self):
        model = StackingClassifier(
            estimators=[
                (
                    "dtc+ceop",
                    CalibratedEqOddsPostprocessing(
                        **self.fairness_info, estimator=DecisionTreeClassifier()
                    ),
                ),
                ("lr", LogisticRegression()),
            ]
        )
        self._attempt_fit_predict(model)

    def test_stacking_post_estimator_mitigation_base_and_final(self):
        model = StackingClassifier(
            estimators=[
                (
                    "dtc+ceop",
                    CalibratedEqOddsPostprocessing(
                        **self.fairness_info, estimator=DecisionTreeClassifier()
                    ),
                ),
                ("lr", LogisticRegression()),
            ],
            final_estimator=CalibratedEqOddsPostprocessing(
                **self.fairness_info, estimator=DecisionTreeClassifier()
            ),
            passthrough=True,
        )
        self._attempt_fit_predict(model)

    def test_stacking_post_estimator_mitigation_final_only(self):
        model = StackingClassifier(
            estimators=[
                ("dtc", DecisionTreeClassifier()),
                ("lr", LogisticRegression()),
            ],
            final_estimator=CalibratedEqOddsPostprocessing(
                **self.fairness_info, estimator=DecisionTreeClassifier()
            ),
            passthrough=True,
        )
        self._attempt_fit_predict(model)



# Copyright 2019, 2020, 2021, 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import importlib
import keyword
import logging
import math
import pprint
import re
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

import black
import numpy as np
import sklearn.metrics

import lale.expressions
import lale.json_operator
import lale.operators
import lale.type_checking

logger = logging.getLogger(__name__)

JSON_TYPE = Dict[str, Any]
_black78 = black.FileMode(line_length=78)


class _CodeGenState:
    imports: List[str]
    assigns: List[str]
    _names: Set[str]

    def __init__(
        self, names: Set[str], combinators: bool, customize_schema: bool, astype: str
    ):
        self.imports = []
        self.assigns = []
        self.combinators = combinators
        self.customize_schema = customize_schema
        self.astype = astype
        self._names = (
            {
                "make_pipeline_graph",
                "lale",
                "make_choice",
                "make_pipeline",
                "make_union",
                "make_union_no_concat",
                "np",
                "pd",
                "pipeline",
            }
            | set(keyword.kwlist)
            | names
        )

    def gensym(self, prefix: str) -> str:
        if prefix in self._names:
            suffix = 0
            while f"{prefix}_{suffix}" in self._names:
                suffix += 1
            result = f"{prefix}_{suffix}"
        else:
            result = prefix
        self._names |= {result}
        return result


def hyperparams_to_string(
    hps: JSON_TYPE, steps: Optional[Dict[str, str]] = None, gen: _CodeGenState = None
) -> str:
    def sklearn_module(value):
        module = value.__module__
        if module.startswith("sklearn."):
            i = module.rfind(".")
            if module[i + 1] == "_":
                module = module[:i]
        return module

    def value_to_string(value):
        if isinstance(value, dict):
            if "$ref" in value and steps is not None:
                step_uid = value["$ref"].split("/")[-1]
                return steps[step_uid]
            else:
                sl = {f"'{k}': {value_to_string(v)}" for k, v in value.items()}
                return "{" + ", ".join(sl) + "}"
        elif isinstance(value, tuple):
            sl = [value_to_string(v) for v in value]
            return "(" + ", ".join(sl) + ")"
        elif isinstance(value, list):
            sl = [value_to_string(v) for v in value]
            return "[" + ", ".join(sl) + "]"
        elif isinstance(value, range):
            return str(value)
        elif isinstance(value, (int, float)) and math.isnan(value):
            return "float('nan')"
        elif isinstance(value, np.dtype):
            if gen is not None:
                gen.imports.append("import numpy as np")
            return f"np.{value.__repr__()}"
        elif isinstance(value, np.ndarray):
            if gen is not None:
                gen.imports.append("import numpy as np")
            array_expr = f"np.{value.__repr__()}"
            # For an array string representation, numpy includes dtype for some data types
            # we need to insert "np." for the dtype so that executing the pretty printed code
            # does not give any error for the dtype. The following code manipulates the
            # string representation given by numpy to add "np." for dtype.
            dtype_indx = array_expr.find("dtype")
            if dtype_indx != -1:
                array_dtype_expr = array_expr[dtype_indx:]
                dtype_name = array_dtype_expr.split("=")[1]
                return array_expr[:dtype_indx] + "dtype=np." + dtype_name
            return array_expr
        elif isinstance(value, np.ufunc):
            if gen is not None:
                gen.imports.append("import numpy as np")
            return f"np.{value.__name__}"  # type: ignore
        elif isinstance(value, lale.expressions.Expr):
            v: lale.expressions.Expr = value
            e = v._expr
            if gen is not None:
                gen.imports.append("from lale.expressions import it")
                for node in ast.walk(e):
                    if isinstance(node, ast.Call):
                        f: Any = node.func
                        gen.imports.append("from lale.expressions import " + f.id)
            return str(value)
        elif hasattr(value, "__module__") and hasattr(value, "__name__"):
            modules = {"numpy": "np", "pandas": "pd"}
            module = modules.get(value.__module__, value.__module__)
            if gen is not None:
                if value.__module__ == module:
                    gen.imports.append(f"import {module}")
                else:
                    gen.imports.append(f"import {value.__module__} as {module}")
            return f"{module}.{value.__name__}"  # type: ignore
        elif hasattr(value, "get_params"):
            module = sklearn_module(value)
            if gen is not None:
                gen.imports.append(f"import {module}")
            actuals = value.get_params(False)  # type: ignore
            defaults = lale.type_checking.get_hyperparam_defaults(value)
            non_defaults = {
                k: v
                for k, v in actuals.items()
                if k not in defaults or defaults[k] != v
            }
            kwargs_string = hyperparams_to_string(non_defaults, steps, gen)
            printed = f"{module}.{value.__class__.__name__}({kwargs_string})"
            return printed
        elif hasattr(sklearn.metrics, "_scorer") and isinstance(
            value, sklearn.metrics._scorer._BaseScorer
        ):
            if gen is not None:
                gen.imports.append("import sklearn.metrics")
            func = value._score_func  # type: ignore
            module = sklearn_module(func)
            if gen is not None:
                gen.imports.append(f"import {module}")
            func_string = f"{module}.{func.__name__}"
            sign_strings = [] if value._sign > 0 else ["greater_is_better=False"]  # type: ignore
            kwargs_strings = [
                f"{k}={value_to_string(v)}" for k, v in value._kwargs.items()  # type: ignore
            ]
            args_strings = [func_string, *sign_strings, *kwargs_strings]
            printed = f"sklearn.metrics.make_scorer({', '.join(args_strings)})"
            return printed
        else:
            printed = pprint.pformat(value, width=10000, compact=True)
            if printed.endswith(")"):
                m = re.match(r"(\w+)\(", printed)
                if m:
                    module = value.__module__
                    if gen is not None:
                        gen.imports.append(f"import {module}")
                    printed = f"{module}.{printed}"
            if printed.startswith("<"):
                m = re.match(r"<(\w[\w.]*)\.(\w+) object at 0x[0-9a-fA-F]+>$", printed)
                if m:
                    module, clazz = m.group(1), m.group(2)
                    if gen is not None:
                        gen.imports.append(f"import {module}")
                    # logger.warning(f"bare {clazz} with unknown constructor")
                    printed = f"{module}.{clazz}()"
            return printed

    strings = [f"{k}={value_to_string(v)}" for k, v in hps.items()]
    return ", ".join(strings)


def _get_module_name(op_label: str, op_name: str, class_name: str) -> str:
    def find_op(module_name, sym):
        module = importlib.import_module(module_name)
        if hasattr(module, sym):
            op = getattr(module, sym)
            if isinstance(op, lale.operators.IndividualOp):
                if op.class_name() == class_name:
                    return op
            elif hasattr(op, "__init__") and hasattr(op, "fit"):
                if hasattr(op, "predict") or hasattr(op, "transform"):
                    return op
        return None

    mod_name_long = class_name[: class_name.rfind(".")]
    if mod_name_long.rfind(".") == -1:
        mod_name_short = mod_name_long
    else:
        mod_name_short = mod_name_long[: mod_name_long.rfind(".")]
    unqualified = class_name[class_name.rfind(".") + 1 :]
    if (
        class_name.startswith("lale.")
        and unqualified.startswith("_")
        and unqualified.endswith("Impl")
    ):
        unqualified = unqualified[1 : -len("Impl")]
    op = find_op(mod_name_short, op_name)
    if op is not None:
        mod = mod_name_short
    else:
        op = find_op(mod_name_long, op_name)
        if op is not None:
            mod = mod_name_long
        else:
            op = find_op(mod_name_short, unqualified)
            if op is not None:
                mod = mod_name_short
            else:
                op = find_op(mod_name_long, unqualified)
                if op is not None:
                    mod = mod_name_long
                else:
                    assert False, (op_label, op_name, class_name)

    assert op is not None, (op_label, op_name, class_name)
    if isinstance(op, lale.operators.IndividualOp):
        if "import_from" in op._schemas:
            mod = op._schemas["import_from"]
    return mod


def _op_kind(op: JSON_TYPE) -> str:
    assert isinstance(op, dict)
    if "kind" in op:
        return op["kind"]
    return lale.json_operator.json_op_kind(op)


_OP_KIND_TO_COMBINATOR = {"Seq": ">>", "Par": "&", "OperatorChoice": "|"}
_OP_KIND_TO_FUNCTION = {
    "Seq": "make_pipeline",
    "Par": "make_union_no_concat",
    "OperatorChoice": "make_choice",
    "Union": "make_union",
}


def _introduce_structure(pipeline: JSON_TYPE, gen: _CodeGenState) -> JSON_TYPE:
    assert _op_kind(pipeline) == "Pipeline"

    def make_graph(pipeline: JSON_TYPE) -> JSON_TYPE:
        steps = pipeline["steps"]
        preds: Dict[str, List[str]] = {step: [] for step in steps}
        succs: Dict[str, List[str]] = {step: [] for step in steps}
        for (src, dst) in pipeline["edges"]:
            preds[dst].append(src)
            succs[src].append(dst)
        return {"kind": "Graph", "steps": steps, "preds": preds, "succs": succs}

    def find_seq(
        graph: JSON_TYPE,
    ) -> Optional[Tuple[Dict[str, JSON_TYPE], Dict[str, JSON_TYPE]]]:
        for src in graph["steps"]:
            if len(graph["succs"][src]) == 1:
                dst = graph["succs"][src][0]
                if len(graph["preds"][dst]) == 1:
                    old: Dict[str, JSON_TYPE] = {
                        uid: graph["steps"][uid] for uid in [src, dst]
                    }
                    new_uid = None
                    new_steps: Dict[str, JSON_TYPE] = {}
                    for step_uid, step_jsn in old.items():
                        if _op_kind(step_jsn) == "Seq":  # flatten
                            new_steps.update(step_jsn["steps"])
                            if new_uid is None:
                                new_uid = step_uid
                        else:
                            new_steps[step_uid] = step_jsn
                    if new_uid is None:
                        new_uid = gen.gensym("pipeline")
                    new = {new_uid: {"kind": "Seq", "steps": new_steps}}
                    return old, new
        return None

    def find_par(
        graph: JSON_TYPE,
    ) -> Optional[Tuple[Dict[str, JSON_TYPE], Dict[str, JSON_TYPE]]]:
        step_uids = list(graph["steps"].keys())
        for i0 in range(len(step_uids)):
            for i1 in range(i0 + 1, len(step_uids)):
                s0, s1 = step_uids[i0], step_uids[i1]
                preds0, preds1 = graph["preds"][s0], graph["preds"][s1]
                if len(preds0) == len(preds1) and set(preds0) == set(preds1):
                    succs0, succs1 = graph["succs"][s0], graph["succs"][s1]
                    if len(succs0) == len(succs1) and set(succs0) == set(succs1):
                        old: Dict[str, JSON_TYPE] = {
                            uid: graph["steps"][uid] for uid in [s0, s1]
                        }
                        new_uid = None
                        new_steps: Dict[str, JSON_TYPE] = {}
                        for step_uid, step_jsn in old.items():
                            if _op_kind(step_jsn) == "Par":  # flatten
                                new_steps.update(step_jsn["steps"])
                                if new_uid is None:
                                    new_uid = step_uid
                            else:
                                new_steps[step_uid] = step_jsn
                        if new_uid is None:
                            new_uid = gen.gensym("union")
                        new: Dict[str, JSON_TYPE] = {
                            new_uid: {"kind": "Par", "steps": new_steps}
                        }
                        return old, new
        return None

    def find_union(
        graph: JSON_TYPE,
    ) -> Optional[Tuple[Dict[str, JSON_TYPE], Dict[str, JSON_TYPE]]]:
        cat_cls = "lale.lib.lale.concat_features._ConcatFeaturesImpl"
        for seq_uid, seq_jsn in graph["steps"].items():
            if _op_kind(seq_jsn) == "Seq":
                seq_uids = list(seq_jsn["steps"].keys())
                for i in range(len(seq_uids) - 1):
                    src, dst = seq_uids[i], seq_uids[i + 1]
                    src_jsn = seq_jsn["steps"][src]
                    if _op_kind(src_jsn) == "Par":
                        dst_jsn = seq_jsn["steps"][dst]
                        if dst_jsn.get("class", None) == cat_cls:
                            old = {seq_uid: seq_jsn}
                            union = {"kind": "Union", "steps": src_jsn["steps"]}
                            if len(seq_uids) == 2:
                                new = {src: union}
                            else:
                                new_steps: Dict[str, JSON_TYPE] = {}
                                for uid, jsn in seq_jsn["steps"].items():
                                    if uid == src:
                                        new_steps[uid] = union
                                    elif uid != dst:
                                        new_steps[uid] = jsn
                                new = {src: {"kind": "Seq", "steps": new_steps}}
                            return old, new
        return None

    def replace(
        subject: JSON_TYPE, old: Dict[str, JSON_TYPE], new: Dict[str, JSON_TYPE]
    ) -> JSON_TYPE:
        assert _op_kind(subject) == "Graph"
        new_uid, new_jsn = list(new.items())[0]
        assert _op_kind(new_jsn) in ["Seq", "Par", "Union"]
        subj_steps = subject["steps"]
        subj_preds = subject["preds"]
        subj_succs = subject["succs"]
        res_steps: Dict[str, JSON_TYPE] = {}
        res_preds: Dict[str, List[str]] = {}
        res_succs: Dict[str, List[str]] = {}
        old_steps_uids = list(old.keys())
        for step_uid in subj_steps:  # careful to keep topological order
            if step_uid == old_steps_uids[0]:
                res_steps[new_uid] = new_jsn
                res_preds[new_uid] = subj_preds[old_steps_uids[0]]
                res_succs[new_uid] = subj_succs[old_steps_uids[-1]]
            elif step_uid not in old_steps_uids:
                res_steps[step_uid] = subj_steps[step_uid]
                res_preds[step_uid] = []
                for pred in subj_preds[step_uid]:
                    if pred == old_steps_uids[-1]:
                        res_preds[step_uid].append(new_uid)
                    elif pred not in old_steps_uids:
                        res_preds[step_uid].append(pred)
                res_succs[step_uid] = []
                for succ in subj_succs[step_uid]:
                    if succ == old_steps_uids[0]:
                        res_succs[step_uid].append(new_uid)
                    elif succ not in old_steps_uids:
                        res_succs[step_uid].append(succ)
        result = {
            "kind": "Graph",
            "steps": res_steps,
            "preds": res_preds,
            "succs": res_succs,
        }
        return result

    def find_and_replace(graph: JSON_TYPE) -> JSON_TYPE:
        if len(graph["steps"]) == 1:  # singleton
            return {"kind": "Seq", "steps": graph["steps"]}
        progress = True
        while progress:
            seq = find_seq(graph)
            if seq is not None:
                graph = replace(graph, *seq)
            par = find_par(graph)
            if par is not None:
                graph = replace(graph, *par)
            if not gen.combinators:
                union = find_union(graph)
                if union is not None:
                    graph = replace(graph, *union)
            progress = seq is not None or par is not None
        if len(graph["steps"]) == 1:  # flatten
            return list(graph["steps"].values())[0]
        else:
            return graph

    graph = make_graph(pipeline)
    result = find_and_replace(graph)
    return result


def _operator_jsn_to_string_rec(uid: str, jsn: JSON_TYPE, gen: _CodeGenState) -> str:
    op_expr: str
    if _op_kind(jsn) == "Pipeline":
        structured = _introduce_structure(jsn, gen)
        return _operator_jsn_to_string_rec(uid, structured, gen)
    elif _op_kind(jsn) == "Graph":
        steps, succs = jsn["steps"], jsn["succs"]
        step2name: Dict[str, str] = {}
        for step_uid, step_val in steps.items():
            expr = _operator_jsn_to_string_rec(step_uid, step_val, gen)
            if re.fullmatch("[A-Za-z][A-Za-z0-9_]*", expr):
                step2name[step_uid] = expr
            else:
                step2name[step_uid] = step_uid
                gen.assigns.append(f"{step_uid} = {expr}")
        make_pipeline = "make_pipeline_graph"
        gen.imports.append(f"from lale.operators import {make_pipeline}")
        result = "{}(steps=[{}], edges=[{}])".format(
            make_pipeline,
            ", ".join([step2name[step] for step in steps]),
            ", ".join(
                [
                    f"({step2name[src]},{step2name[tgt]})"
                    for src in steps
                    for tgt in succs[src]
                ]
            ),
        )
        return result
    elif _op_kind(jsn) in ["Seq", "Par", "OperatorChoice", "Union"]:
        if gen.combinators:

            def print_for_comb(step_uid, step_val):
                printed = _operator_jsn_to_string_rec(step_uid, step_val, gen)
                parens = _op_kind(step_val) != _op_kind(jsn) and _op_kind(step_val) in [
                    "Seq",
                    "Par",
                    "OperatorChoice",
                ]
                return f"({printed})" if parens else printed

            printed_steps = {
                step_uid: print_for_comb(step_uid, step_val)
                for step_uid, step_val in jsn["steps"].items()
            }
            combinator = _OP_KIND_TO_COMBINATOR[_op_kind(jsn)]
            if len(printed_steps.values()) == 1 and combinator == ">>":
                gen.imports.append("from lale.operators import make_pipeline")
                op_expr = "make_pipeline({})".format(", ".join(printed_steps.values()))
                return op_expr
            return f" {combinator} ".join(printed_steps.values())
        else:
            printed_steps = {
                step_uid: _operator_jsn_to_string_rec(step_uid, step_val, gen)
                for step_uid, step_val in jsn["steps"].items()
            }
            function = _OP_KIND_TO_FUNCTION[_op_kind(jsn)]
            if gen.astype == "sklearn" and function in ["make_union", "make_pipeline"]:
                gen.imports.append(f"from sklearn.pipeline import {function}")
            else:
                gen.imports.append(f"from lale.operators import {function}")
            op_expr = "{}({})".format(function, ", ".join(printed_steps.values()))
            gen.assigns.append(f"{uid} = {op_expr}")
            return uid
    elif _op_kind(jsn) == "IndividualOp":
        label: str = jsn["label"]
        class_name = jsn["class"]
        module_name = _get_module_name(label, jsn["operator"], class_name)
        if module_name.startswith("lale."):
            op_name = jsn["operator"]
        else:
            op_name = class_name[class_name.rfind(".") + 1 :]
        if op_name.startswith("_"):
            op_name = op_name[1:]
        if op_name.endswith("Impl"):
            op_name = op_name[: -len("Impl")]
        if op_name == label:
            import_stmt = f"from {module_name} import {op_name}"
        else:
            import_stmt = f"from {module_name} import {op_name} as {label}"
        gen.imports.append(import_stmt)
        printed_steps = {
            step_uid: _operator_jsn_to_string_rec(step_uid, step_val, gen)
            for step_uid, step_val in jsn.get("steps", {}).items()
        }
        op_expr = label
        if "customize_schema" in jsn and gen.customize_schema:
            if jsn["customize_schema"] == "not_available":
                logger.warning(f"missing {label}.customize_schema(..) call")
            elif jsn["customize_schema"] != {}:
                new_hps = lale.json_operator._top_schemas_to_hp_props(
                    jsn["customize_schema"]
                )
                customize_schema_string = ",".join(
                    [
                        f"{hp_name}={json_to_string(hp_schema)}"
                        for hp_name, hp_schema in new_hps.items()
                    ]
                )
                op_expr = f"{op_expr}.customize_schema({customize_schema_string})"
        if "hyperparams" in jsn and jsn["hyperparams"] is not None:
            hp_string = hyperparams_to_string(jsn["hyperparams"], printed_steps, gen)
            op_expr = f"{op_expr}({hp_string})"
        if re.fullmatch(r".+\(.+\)", op_expr):
            gen.assigns.append(f"{uid} = {op_expr}")
            return uid
        else:
            return op_expr
    else:
        assert False, f"unexpected type {type(jsn)} of jsn {jsn}"


def _collect_names(jsn: JSON_TYPE) -> Set[str]:
    result: Set[str] = set()
    if "steps" in jsn:
        steps: Dict[str, JSON_TYPE] = jsn["steps"]
        for step_uid, step_jsn in steps.items():
            result |= {step_uid}
            result |= _collect_names(step_jsn)
    if "label" in jsn:
        lbl: str = jsn["label"]
        result |= {lbl}
    return result


def _combine_lonely_literals(printed_code):
    lines = printed_code.split("\n")
    regex = re.compile(
        r' +("[^"]*"|\d+\.?\d*|\[\]|float\("nan"\)|np\.dtype\("[^"]+"\)),'
    )
    for i in range(len(lines)):
        if lines[i] is not None:
            match_i = regex.fullmatch(lines[i])
            if match_i is not None:
                j = i + 1
                while j < len(lines) and lines[j] is not None:
                    match_j = regex.fullmatch(lines[j])
                    if match_j is None:
                        break
                    candidate = lines[i] + " " + match_j.group(1) + ","
                    if len(candidate) > 78:
                        break
                    lines[i] = candidate
                    lines[j] = None
                    j += 1
    result = "\n".join([s for s in lines if s is not None])
    return result


def _format_code(printed_code):
    formatted = black.format_str(printed_code, mode=_black78).rstrip()
    combined = _combine_lonely_literals(formatted)
    return combined


def _operator_jsn_to_string(
    jsn: JSON_TYPE,
    show_imports: bool,
    combinators: bool,
    customize_schema: bool,
    astype: str,
) -> str:
    gen = _CodeGenState(_collect_names(jsn), combinators, customize_schema, astype)
    expr = _operator_jsn_to_string_rec("pipeline", jsn, gen)
    if expr != "pipeline":
        gen.assigns.append(f"pipeline = {expr}")
    if show_imports and len(gen.imports) > 0:
        if combinators:
            gen.imports.append("import lale")
        imports_set: Set[str] = set()
        imports_list: List[str] = []
        for imp in gen.imports:
            if imp not in imports_set:
                imports_set |= {imp}
                imports_list.append(imp)
        result = "\n".join(imports_list)
        if combinators:
            result += "\nlale.wrap_imported_operators()"
        result += "\n"
        result += "\n".join(gen.assigns)
    else:
        result = "\n".join(gen.assigns)
    formatted = _format_code(result)
    return formatted


def json_to_string(jsn: JSON_TYPE) -> str:
    def _inner(value):
        if value is None:
            return "None"
        elif isinstance(value, (bool, str)):
            return pprint.pformat(value, width=10000, compact=True)
        elif isinstance(value, (int, float)):
            if math.isnan(value):
                return "float('nan')"
            else:
                return pprint.pformat(value, width=10000, compact=True)
        elif isinstance(value, list):
            sl = [_inner(v) for v in value]
            return "[" + ", ".join(sl) + "]"
        elif isinstance(value, tuple):
            sl = [_inner(v) for v in value]
            return "(" + ", ".join(sl) + ")"
        elif isinstance(value, dict):
            sl = [f"'{k}': {_inner(v)}" for k, v in value.items()]
            return "{" + ", ".join(sl) + "}"
        else:
            return f"<<{type(value).__qualname__}>>"

    s1 = _inner(jsn)
    s2 = _format_code(s1)
    return s2


def to_string(
    arg: Union[JSON_TYPE, "lale.operators.Operator"],
    show_imports: bool = True,
    combinators: bool = True,
    customize_schema: bool = False,
    astype: str = "lale",
    call_depth: int = 1,
) -> str:
    assert astype in ["lale", "sklearn"], astype
    if astype == "sklearn":
        combinators = False
    if lale.type_checking.is_schema(arg):
        return json_to_string(cast(JSON_TYPE, arg))
    elif isinstance(arg, lale.operators.Operator):
        jsn = lale.json_operator.to_json(
            arg,
            call_depth=call_depth + 1,
            add_custom_default=not customize_schema,
        )
        return _operator_jsn_to_string(
            jsn, show_imports, combinators, customize_schema, astype
        )
    else:
        raise ValueError(f"Unexpected argument type {type(arg)} for {arg}")


def ipython_display(
    arg: Union[JSON_TYPE, "lale.operators.Operator"],
    show_imports: bool = True,
    combinators: bool = True,
):
    import IPython.display

    pretty_printed = to_string(arg, show_imports, combinators, call_depth=3)
    markdown = IPython.display.Markdown(f"```python\n{pretty_printed}\n```")
    IPython.display.display(markdown)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.svm

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": [
                "kernel",
                "degree",
                "gamma",
                "coef0",
                "tol",
                "C",
                "epsilon",
                "shrinking",
                "cache_size",
                "verbose",
                "max_iter",
            ],
            "relevantToOptimizer": [
                "kernel",
                "degree",
                "gamma",
                "C",
                "shrinking",
                "tol",
            ],
            "properties": {
                "kernel": {
                    "anyOf": [
                        {"enum": ["precomputed"], "forOptimizer": False},
                        {"enum": ["linear", "poly", "rbf", "sigmoid"]},
                        {"laleType": "callable", "forOptimizer": False},
                    ],
                    "default": "rbf",
                    "description": "Specifies the kernel type to be used in the algorithm.",
                },
                "degree": {
                    "type": "integer",
                    "minimum": 0,
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 5,
                    "default": 3,
                    "description": "Degree of the polynomial kernel function ('poly').",
                },
                "gamma": {
                    "anyOf": [
                        {
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                            "minimumForOptimizer": 3.0517578125e-05,
                            "maximumForOptimizer": 8,
                            "distribution": "loguniform",
                        },
                        {"enum": ["auto", "auto_deprecated", "scale"]},
                    ],
                    "default": "auto_deprecated",  # going to change to 'scale' from sklearn 0.22.
                    "description": "Kernel coefficient for 'rbf', 'poly', and 'sigmoid'.",
                },
                "coef0": {
                    "type": "number",
                    "minimumForOptimizer": -1,
                    "maximumForOptimizer": 1,
                    "default": 0.0,
                    "description": "Independent term in kernel function.",
                },
                "tol": {
                    "type": "number",
                    "distribution": "loguniform",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "maximumForOptimizer": 0.01,
                    "default": 0.001,
                    "description": "Tolerance for stopping criteria.",
                },
                "C": {
                    "description": "Penalty parameter C of the error term.",
                    "type": "number",
                    "distribution": "loguniform",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "default": 1.0,
                    "minimumForOptimizer": 0.03125,
                    "maximumForOptimizer": 32768,
                },
                "epsilon": {
                    "description": "Epsilon in the epsilon-SVR model. It specifies the epsilon-tube within which no penalty is associated in the training loss function with points predicted within a distance epsilon from the actual value.",
                    "type": "number",
                    "default": 0.1,
                    "minimum": 0.0,
                    "minimumForOptimizer": 0.00001,
                    "maximumForOptimizer": 10000.0,
                },
                "shrinking": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to use the shrinking heuristic.",
                },
                "cache_size": {
                    "type": "number",
                    "minimum": 0,
                    "maximumForOptimizer": 1000,
                    "default": 200.0,
                    "description": "Specify the size of the kernel cache (in MB).",
                },
                "verbose": {
                    "type": "boolean",
                    "default": False,
                    "description": "Enable verbose output.",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 1,
                    "maximumForOptimizer": 1000,
                    "default": -1,
                    "description": "Hard limit on iterations within solver, or -1 for no limit.",
                },
            },
        },
        {
            "description": "Sparse precomputed kernels are not supported.",
            "anyOf": [
                {"type": "object", "laleNot": "X/isSparse"},
                {
                    "type": "object",
                    "properties": {"kernel": {"not": {"enum": ["precomputed"]}}},
                },
            ],
        },
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "y": {"type": "array", "items": {"type": "number"}},
    },
}

_input_predict_schema = {
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        }
    },
}

_output_predict_schema = {
    "description": "The predicted classes.",
    "type": "array",
    "items": {"type": "number"},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Support Vector Classification`_ from scikit-learn.

.. _`Support Vector Classification`: https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.svc.html",
    "import_from": "sklearn.svm",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}

SVR: lale.operators.PlannedIndividualOp
SVR = lale.operators.make_operator(sklearn.svm.SVR, _combined_schemas)

if sklearn.__version__ >= "0.22":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.svm.SVR.html
    # new: https://scikit-learn.org/0.23/modules/generated/sklearn.svm.SVR.html
    from lale.schemas import AnyOf, Enum, Float

    SVR = SVR.customize_schema(
        gamma=AnyOf(
            types=[
                Enum(["scale", "auto"]),
                Float(
                    minimum=0.0,
                    exclusiveMinimum=True,
                    minimumForOptimizer=3.0517578125e-05,
                    maximumForOptimizer=8,
                    distribution="loguniform",
                ),
            ],
            desc="Kernel coefficient for 'rbf', 'poly' and 'sigmoid'.",
            default="scale",
        ),
        set_as_available=True,
    )


lale.docstrings.set_docstrings(SVR)



# Copyright 2020, 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aif360.algorithms.inprocessing
import sklearn.linear_model

import lale.docstrings
import lale.operators

from .util import (
    _BaseInEstimatorImpl,
    _categorical_fairness_properties,
    _categorical_input_predict_proba_schema,
    _categorical_input_predict_schema,
    _categorical_output_predict_proba_schema,
    _categorical_output_predict_schema,
    _categorical_supervised_input_fit_schema,
)


class _GerryFairClassifierImpl(_BaseInEstimatorImpl):
    def __init__(
        self,
        *,
        favorable_labels,
        protected_attributes,
        unfavorable_labels=None,
        redact=True,
        preparation=None,
        C=10,
        printflag=False,
        heatmapflag=False,
        heatmap_iter=10,
        heatmap_path=".",
        max_iters=10,
        gamma=0.01,
        fairness_def="FP",
        predictor=None,
    ):
        if predictor is None:
            predictor = sklearn.linear_model.LinearRegression()
        if isinstance(predictor, lale.operators.Operator):
            if isinstance(predictor, lale.operators.IndividualOp):
                predictor = predictor._impl_instance()._wrapped_model
            else:
                raise ValueError(
                    "If predictor is a Lale operator, it needs to be an individual operator."
                )
        mitigator = aif360.algorithms.inprocessing.GerryFairClassifier(
            C=C,
            printflag=printflag,
            heatmapflag=heatmapflag,
            heatmap_iter=heatmap_iter,
            heatmap_path=heatmap_path,
            max_iters=max_iters,
            gamma=gamma,
            fairness_def=fairness_def,
            predictor=predictor,
        )
        super(_GerryFairClassifierImpl, self).__init__(
            favorable_labels=favorable_labels,
            protected_attributes=protected_attributes,
            unfavorable_labels=unfavorable_labels,
            redact=redact,
            preparation=preparation,
            mitigator=mitigator,
        )


_input_fit_schema = _categorical_supervised_input_fit_schema
_input_predict_schema = _categorical_input_predict_schema
_output_predict_schema = _categorical_output_predict_schema
_input_predict_proba_schema = _categorical_input_predict_proba_schema
_output_predict_proba_schema = _categorical_output_predict_proba_schema

_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                *_categorical_fairness_properties.keys(),
                "redact",
                "preparation",
                "C",
                "printflag",
                "heatmapflag",
                "heatmap_iter",
                "heatmap_path",
                "max_iters",
                "gamma",
                "fairness_def",
                "predictor",
            ],
            "relevantToOptimizer": ["C", "max_iters", "gamma", "fairness_def"],
            "properties": {
                **_categorical_fairness_properties,
                "redact": {
                    "description": "Whether to redact protected attributes before data preparation (recommended) or not.",
                    "type": "boolean",
                    "default": True,
                },
                "preparation": {
                    "description": "Transformer, which may be an individual operator or a sub-pipeline.",
                    "anyOf": [
                        {"laleType": "operator"},
                        {"description": "lale.lib.lale.NoOp", "enum": [None]},
                    ],
                    "default": None,
                },
                "C": {
                    "description": "Maximum L1 norm for the dual variables.",
                    "type": "number",
                    "default": 10,
                    "minimumForOptimizer": 0.03125,
                    "maximumForOptimizer": 32768,
                },
                "printflag": {
                    "description": "Print output flag.",
                    "type": "boolean",
                    "default": False,
                },
                "heatmapflag": {
                    "description": "Save heatmaps every heatmap_iter flag.",
                    "type": "boolean",
                    "default": False,
                },
                "heatmap_iter": {
                    "description": "Save heatmaps every heatmap_iter.",
                    "type": "integer",
                    "minimum": 1,
                    "default": 10,
                },
                "heatmap_path": {
                    "description": "Save heatmaps path.",
                    "type": "string",
                    "default": ".",
                },
                "max_iters": {
                    "description": "Time horizon for the fictitious play dynamic.",
                    "type": "integer",
                    "minimum": 1,
                    "default": 10,
                    "distribution": "loguniform",
                    "maximumForOptimizer": 1000,
                },
                "gamma": {
                    "description": "Fairness approximation parameter.",
                    "type": "number",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "default": 0.01,
                    "distribution": "loguniform",
                    "minimumForOptimizer": 0.001,
                    "maximumForOptimizer": 1.0,
                },
                "fairness_def": {
                    "description": "Fairness notion.",
                    "enum": ["FP", "FN"],
                    "default": "FP",
                },
                "predictor": {
                    "description": "Hypothesis class for the learner.",
                    "anyOf": [
                        {
                            "description": "Supports LR, SVM, KR, Trees.",
                            "laleType": "operator",
                        },
                        {
                            "description": "sklearn.linear_model.LinearRegression",
                            "enum": [None],
                        },
                    ],
                    "default": None,
                },
            },
        },
    ],
}

_combined_schemas = {
    "description": """`GerryFairClassifier`_ in-estimator fairness mitigator. Attempts to learn classifiers that are fair with respect to rich subgroups (`Kearns et al. 2018`_, `Kearns et al. 2019`_). Rich subgroups are defined by (linear) functions over the sensitive attributes, and fairness notions are statistical: false positive, false negative, and statistical parity rates. This implementation uses a max of two regressions as a cost-sensitive classification oracle, and supports linear regression, support vector machines, decision trees, and kernel regression.

.. _`GerryFairClassifier`: https://aif360.readthedocs.io/en/latest/modules/generated/aif360.algorithms.inprocessing.GerryFairClassifier.html
.. _`Kearns et al. 2018`: http://proceedings.mlr.press/v80/kearns18a.html
.. _`Kearns et al. 2019`: https://doi.org/10.1145/3287560.3287592
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.gerry_fair_classifier.html#lale.lib.aif360.gerry_fair_classifier.GerryFairClassifier",
    "import_from": "aif360.sklearn.inprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}

GerryFairClassifier = lale.operators.make_operator(
    _GerryFairClassifierImpl, _combined_schemas
)

lale.docstrings.set_docstrings(GerryFairClassifier)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from autoai_ts_libs.watfore.watfore_forecasters import (  # type: ignore # noqa
    WatForeForecaster as model_to_be_wrapped,
)

import lale.docstrings
import lale.operators


class _WatForeForecasterImpl:
    def __init__(
        self,
        algorithm="hw",
        use_full_error_history=True,
        error_horizon_length=1,
        force_model=False,
        min_training_data=-1,
        p_min=0,
        p_max=-1,
        q_min=0,
        q_max=0,
        algorithm_type="additive",
        samples_per_season=2,
        initial_training_seasons=2,
        compute_seasonality=False,
        error_history_length=1,
        #        training_sample_size=0,
        box_cox_transform=False,
        ts_icol_loc=-1,
        prediction_horizon=1,
        log_transform=True,
        lookback_win=1,
        target_column_indices=-1,
        debug=False,
        **kwargs
    ):
        # code from Wes to fix a specific handling in autoai_ts:
        if "_" in algorithm:
            algorithm_parts = algorithm.split("_")
            algorithm = algorithm_parts[0]
            algorithm_type = algorithm_parts[1]
        self._hyperparams = {
            "algorithm": algorithm,
            "use_full_error_history": use_full_error_history,
            "error_horizon_length": error_horizon_length,
            "force_model": force_model,
            "min_training_data": min_training_data,
            "p_min": p_min,
            "p_max": p_max,
            "q_min": q_min,
            "q_max": q_max,
            "algorithm_type": algorithm_type,
            "samples_per_season": samples_per_season,
            "initial_training_seasons": initial_training_seasons,
            "compute_seasonality": compute_seasonality,
            "error_history_length": error_history_length,
            #            "training_sample_size":training_sample_size,
            "box_cox_transform": box_cox_transform,
            "ts_icol_loc": ts_icol_loc,
            "prediction_horizon": prediction_horizon,
            "log_transform": log_transform,
            "lookback_win": lookback_win,
            "target_column_indices": target_column_indices,
            "debug": debug,
        }
        self._hyperparams.update(**kwargs)
        self._wrapped_model = model_to_be_wrapped(**self._hyperparams)

    def fit(self, X, y):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X=None, **predict_params):
        return self._wrapped_model.predict(X, **predict_params)

    def predict_proba(self, X=None):
        return self._wrapped_model.predict_proba(X)

    def viz_label(self) -> str:
        return "WatForeForecaster_" + self._hyperparams["algorithm"]


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": True,
            "required": [
                "algorithm",
                "use_full_error_history",
                "error_horizon_length",
                "force_model",
                "min_training_data",
                "p_min",
                "p_max",
                "q_min",
                "q_max",
                "algorithm_type",
                "samples_per_season",
                "initial_training_seasons",
                "compute_seasonality",
                "error_history_length",
                #                "training_sample_size",
                "box_cox_transform",
                "ts_icol_loc",
                "prediction_horizon",
                "log_transform",
                "lookback_win",
                "target_column_indices",
                "debug",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "algorithm": {
                    "description": """Algorithm that is used to initialize the prediction model. Currently supported are 'hw' i.e. holtwinters,
'arima','bats', autoforecaster i.e., BATS model with Box-Conx transformation. Algorithm specific parameters
also need to be specified.
Additive and multiplicative variants of the Holt-Winters Seasonal forecasting method. This implementation of
the Holt-Winter algorithm variants assumes that the data it receives satisfy the above conditions.
Any pre-processing the data needs in order to satisfy the above assumptions should take place prior to model
updates and calls for prediction. This approach was followed in order to allow any type of pre-processing
(for example for filling missing values) on the data, independent of the H-W core calculations.
Implementation of BATS (Box-Cox transform, ARMA errors, Trend, and Seasonal components) Algorithms
Reference: Alysha M De Livera, Rob J Hyndman and Ralph D Snyder, "Forecasting time series with complex seasonal
patterns using exponential smoothing," Journal of the American Statistical Association (2011) 106(496), 1513-1527.
If algorithm = autoforecaster, This trains all models and to keep running statistics on their forecasting
errors as updates are requested. The error statistics are used to continually update the selection
of the best model, which is used to do the forecasts in the super class. The algorithm becomes initialized as
soon as the first algorithm becomes initialized so as to allow forecasts as soon as possible. It continues to
rate new algorithms as they become initialized and/or subsequent updates are applied.""",
                    "enum": [
                        "hw",
                        "arima",
                        "bats",
                        "autoforecaster",
                        "hw_additive",
                        "hw_multiplicative",
                    ],
                    "default": "hw",
                },
                "use_full_error_history": {
                    "description": """(ARIMA and HoltWinters (hw) ONLY) Trains arima model using full
error history from the data. If False, then only the last errorHorizonLength updates will be considered in the
values returned. The resulting instance:
1. does not force a model if suitable orders and/or coefficients can not be found. This can result in a model
which can not be initialized.
2.picks the required amount of training data automatically.
3.finds the AR order automatically.
4. finds the MA order automatically.
5.finds the difference order automatically.""",
                    "type": "boolean",
                    "default": True,
                },
                "error_horizon_length": {
                    "description": """(ARIMA ONLY) This parameter is used only when algroithm='arima' or watfore.Forecasters.arima, this is error horizon for
error in arima model.""",
                    "type": "integer",
                    "default": 1,
                },
                "force_model": {
                    "description": """(ARIMA ONLY) If True, then force the selection of a model based on the given orders, regardless of suitability. If False,
then the model may never become initialized if suitable coefficients for the training data can not be identified.""",
                    "type": "boolean",
                    "default": False,
                },
                "min_training_data": {
                    "description": """(ARIMA ONLY) The amount of training data to use to learn the coefficients. May be non-positive, in which case the minimum
amount of training data will be determined by the pMax and qMax values.""",
                    "type": "integer",
                    "default": -1,
                },
                "p_min": {
                    "description": """(ARIMA ONLY) Minimum AR order to be selected during training. Must be 0 or larger.""",
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                },
                "p_max": {
                    "description": """(ARIMA ONLY) Maximum AR order to be selected during training. If less than 0, then the maximum supported order will be used,
otherwise, must be at least as large as pMin.""",
                    "type": "integer",
                    "default": -1,
                },
                "q_min": {
                    "description": """(ARIMA ONLY) Minimum AR order to be selected during training. Must be 0 or larger.""",
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                },
                "q_max": {
                    "description": """(ARIMA ONLY) Maximum MA order to be selected during training. If less than 0, then the maximum supported order will be used,
otherwise, must be at least as large as qMin.""",
                    "type": "integer",
                    "default": 0,
                },
                "algorithm_type": {
                    "description": """(HoltWinters(hw) ONLY, i.e. when algorithm=watfore.Forecasters.hw or algorithm='hw')
`additive` provides implementation of the additive variant of the Holt-Winters Seasonal forecasting method.
The additive variant has the seasonal and trend/slope components enter the forecasting function in an
additive manner (see ref. 1), as in
See http://books.google.com/books?id=GSyzox8Lu9YC&source=gbs_navlinks_s for more information.
y(t+h) = L(t) + t*H(t) + S(t+h)
where
t = latest time for which the model has been updated
h = number of steps ahead for which a forecast is desired
L(t) = is the level estimate at time t
H(t) = is the slope at time t
S(t+h) = is the seasonal component at time t + h.
`multiplicative`, provides implementation of the multiplicative variant of the Holt-Winters Seasonal forecasting
method. The multiplicative variant has the seasonal and trend/slope components enter the forecasting function
in a multiplicative manner (see ref. 1, Brockwell, pp. 329).
y(t+h) = (L(t) + t* H(t)) * S(t+h)""",
                    "enum": ["additive", "multiplicative", None],
                    "default": "additive",
                },
                "samples_per_season": {
                    "description": """(hw ONLY)
Season length, or if compute_seasonality is True, then the maximum season to allow. If used as a maximum season,
then identification of the season length (and thus forecasting) can not happen until samples_per_season
* initial_training_seasons of data are provided.""",
                    "anyOf": [{"type": "integer"}, {"type": "number"}],
                    "default": 2,
                },
                "initial_training_seasons": {
                    "description": """(hw ONLY) Number of seasons to use in training. Must be greater than 1.""",
                    "type": "integer",
                    "default": 2,
                    "minimum": 2,
                },
                "compute_seasonality": {
                    "description": """(hw ONLY) If true then automatically compute seasonality.""",
                    "type": "boolean",
                    "default": False,
                },
                "error_history_length": {
                    "description": """ (hw ONLY)""",
                    "type": "integer",
                    "default": 1,
                },
                #                 "training_sample_size":{
                #                     "description":"""(BATS(bats) ONLY, i.e. when algorithm=watfore.Forecasters.bats or algorithm='bats')
                # training sample size, recommended #samples = 2*maximum_cycle_length""",
                #                     "type":"integer"
                #                 },
                "box_cox_transform": {
                    "description": """(BATS(bats) ONLY, i.e. when algorithm=watfore.Forecasters.bats or algorithm='bats')
Only estimate Box-Cox parameter if True; otherwise, ignore Box-Cox transform""",
                    "type": "boolean",
                    "default": False,
                },
                "ts_icol_loc": {
                    "description": """This parameter tells the forecasting model the absolute location of the timestamp column. For specifying
time stamp location put value in array e.g., [0] if 0th column is time stamp. The array is to support
multiple timestamps in future. If ts_icol_loc = -1 that means no timestamp is provided and all data is
time series. With ts_icol_loc=-1, the model will assume all the data is ordered and equally sampled.""",
                    "anyOf": [
                        {"type": "array", "items": {"type": "integer"}},
                        {"enum": [-1]},
                    ],
                    "default": -1,
                },
                "prediction_horizon": {
                    "description": "The number of time points to predict into the future.",
                    "type": "integer",
                    "default": 1,
                },
                "log_transform": {
                    "description": "Whether a log transform of the data should be applied before fitting the estimator.",
                    "anyOf": [{"type": "boolean"}, {"enum": [None]}],
                    "default": True,
                },
                "lookback_win": {
                    "description": "Not currently utilized.",
                    "type": "integer",
                    "default": 1,
                },
                "target_column_indices": {
                    "description": """This parameter tells the forecasting model the absolute location of the target column(s) that need to be
used for training model(s). While fiting the specified column(s) are used for training and are predicted
subsequently in the predict function. Default is -1 which will assume all columns except timestamp
are targets.""",
                    "anyOf": [
                        {"type": "array", "items": {"type": "integer"}},
                        {"enum": [-1]},
                    ],
                    "default": -1,
                },
                "debug": {"description": """""", "type": "boolean", "default": False},
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        },
        "y": {"laleType": "Any"},
    },
}

_input_predict_schema = {
    "type": "object",
    "additionalProperties": False,
    "required": ["X"],
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"enum": [None]},
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        }
    },
}

_output_predict_schema = {
    "description": "Features; the outer array is over samples.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
    ],
}

_input_predict_proba_schema = {
    "type": "object",
    "additionalProperties": False,
    "required": ["X"],
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"enum": [None]},
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        }
    },
}

_output_predict_proba_schema = {
    "description": "Features; the outer array is over samples.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Operator from `autoai_ts_libs`_.

.. _`autoai_ts_libs`: https://pypi.org/project/autoai-ts-libs""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_ts_libs.mt2r_forecaster.html",
    "import_from": "autoai_ts_libs.watfore.watfore_forecasters",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "forecaster"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}

WatForeForecaster = lale.operators.make_operator(
    _WatForeForecasterImpl, _combined_schemas
)
lale.docstrings.set_docstrings(WatForeForecaster)



# Copyright 2019, 2020, 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aif360.algorithms.preprocessing
import aif360.datasets
import numpy as np
import pandas as pd

import lale.docstrings
import lale.lib.lale
import lale.operators

from .protected_attributes_encoder import ProtectedAttributesEncoder
from .redacting import Redacting
from .util import (
    _categorical_fairness_properties,
    _categorical_input_transform_schema,
    _categorical_supervised_input_fit_schema,
    _numeric_output_transform_schema,
    _validate_fairness_info,
)


class _DisparateImpactRemoverImpl:
    def __init__(
        self,
        *,
        favorable_labels,
        protected_attributes,
        unfavorable_labels=None,
        redact=True,
        preparation=None,
        repair_level=1.0,
    ):
        _validate_fairness_info(
            favorable_labels, protected_attributes, unfavorable_labels, False
        )
        self.favorable_labels = favorable_labels
        self.protected_attributes = protected_attributes
        self.unfavorable_labels = unfavorable_labels
        self.redact = redact
        if preparation is None:
            preparation = lale.lib.lale.NoOp
        self.preparation = preparation
        self.repair_level = repair_level

    def _prep_and_encode(self, X, y=None):
        prepared_X = self.redact_and_prep.transform(X, y)
        encoded_X, encoded_y = self.prot_attr_enc.transform(X, y)
        assert isinstance(encoded_X, pd.DataFrame), type(encoded_X)
        assert encoded_X.shape[1] == 1, encoded_X.columns
        if isinstance(prepared_X, pd.DataFrame):
            combined_attribute_names = list(prepared_X.columns) + [
                name for name in encoded_X.columns if name not in prepared_X.columns
            ]
            combined_columns = [
                encoded_X[name] if name in encoded_X else prepared_X[name]
                for name in combined_attribute_names
            ]
            combined_X = pd.concat(combined_columns, axis=1)
            sensitive_attribute = list(encoded_X.columns)[0]
        else:
            if isinstance(prepared_X, pd.DataFrame):
                prepared_X = prepared_X.to_numpy()
            assert isinstance(prepared_X, np.ndarray)
            encoded_X = encoded_X.to_numpy()
            assert isinstance(encoded_X, np.ndarray)
            combined_X = np.concatenate([prepared_X, encoded_X], axis=1)
            sensitive_attribute = combined_X.shape[1] - 1
        return combined_X, sensitive_attribute

    def fit(self, X, y=None):
        fairness_info = {
            "favorable_labels": self.favorable_labels,
            "protected_attributes": self.protected_attributes,
            "unfavorable_labels": self.unfavorable_labels,
        }
        if self.redact:
            redacting = Redacting(**fairness_info)
            trainable_redact_and_prep = redacting >> self.preparation
        else:
            trainable_redact_and_prep = self.preparation
        assert isinstance(trainable_redact_and_prep, lale.operators.TrainablePipeline)
        self.redact_and_prep = trainable_redact_and_prep.fit(X, y)
        self.prot_attr_enc = ProtectedAttributesEncoder(
            **fairness_info, remainder="drop", return_X_y=True, combine="and"
        )
        encoded_X, sensitive_attribute = self._prep_and_encode(X, y)
        if isinstance(sensitive_attribute, str):
            assert isinstance(encoded_X, pd.DataFrame)
            features = encoded_X.to_numpy().tolist()
            index = encoded_X.columns.to_list().index(sensitive_attribute)
        else:
            assert isinstance(encoded_X, np.ndarray)
            features = encoded_X.tolist()
            index = sensitive_attribute
        # since DisparateImpactRemover does not have separate fit and transform
        di_remover = aif360.algorithms.preprocessing.DisparateImpactRemover(
            repair_level=self.repair_level, sensitive_attribute=sensitive_attribute
        )
        self.mitigator = di_remover.Repairer(features, index, self.repair_level, False)
        return self

    def transform(self, X):
        encoded_X, _ = self._prep_and_encode(X)
        columns = None
        if isinstance(encoded_X, pd.DataFrame):
            features = encoded_X.to_numpy().tolist()
            columns = encoded_X.columns
        else:
            assert isinstance(encoded_X, np.ndarray)
            features = encoded_X.tolist()
        mitigated_X = self.mitigator.repair(features)
        if isinstance(encoded_X, pd.DataFrame):
            result = pd.DataFrame(mitigated_X, index=X.index, columns=columns)
        else:
            result = np.array(mitigated_X)
        return result


_input_fit_schema = _categorical_supervised_input_fit_schema
_input_transform_schema = _categorical_input_transform_schema
_output_transform_schema = _numeric_output_transform_schema

_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": [
                *_categorical_fairness_properties.keys(),
                "redact",
                "preparation",
                "repair_level",
            ],
            "relevantToOptimizer": ["repair_level"],
            "properties": {
                **_categorical_fairness_properties,
                "redact": {
                    "description": "Whether to redact protected attributes before data preparation (recommended) or not.",
                    "type": "boolean",
                    "default": True,
                },
                "preparation": {
                    "description": "Transformer, which may be an individual operator or a sub-pipeline.",
                    "anyOf": [
                        {"laleType": "operator"},
                        {"description": "lale.lib.lale.NoOp", "enum": [None]},
                    ],
                    "default": None,
                },
                "repair_level": {
                    "description": "Repair amount from 0 = none to 1 = full.",
                    "type": "number",
                    "minimum": 0,
                    "maximum": 1,
                    "default": 1,
                },
            },
        }
    ],
}

_combined_schemas = {
    "description": """`Disparate impact remover`_ pre-estimator fairness mitigator. Edits feature values to increase group fairness while preserving rank-ordering within groups (`Feldman et al. 2015`_). In the case of multiple protected attributes, the combined reference group is the intersection of the reference groups for each attribute.

.. _`Disparate impact remover`: https://aif360.readthedocs.io/en/latest/modules/generated/aif360.algorithms.preprocessing.DisparateImpactRemover.html
.. _`Feldman et al. 2015`: https://doi.org/10.1145/2783258.2783311
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.disparate_impact_remover.html#lale.lib.aif360.disparate_impact_remover.DisparateImpactRemover",
    "import_from": "aif360.algorithms.preprocessing",
    "type": "object",
    "tags": {"pre": ["~categoricals"], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

DisparateImpactRemover = lale.operators.make_operator(
    _DisparateImpactRemoverImpl, _combined_schemas
)

lale.docstrings.set_docstrings(DisparateImpactRemover)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from lale.lib.lale import NoOp
from lale.lib.sklearn import (
    PCA,
    RFE,
    AdaBoostRegressor,
    DecisionTreeClassifier,
    LinearRegression,
    LogisticRegression,
    SelectKBest,
    SimpleImputer,
    StackingClassifier,
    VotingClassifier,
)


class TestReplace(unittest.TestCase):
    def test_choice(self):
        two_choice = PCA | SelectKBest
        replaced_choice = two_choice.replace(PCA, LogisticRegression)
        expected_choice = LogisticRegression | SelectKBest
        self.assertEqual(replaced_choice.to_json(), expected_choice.to_json())

        three_choice = PCA | SelectKBest | NoOp
        replaced_choice = three_choice.replace(PCA, LogisticRegression)
        expected_choice = LogisticRegression | SelectKBest | NoOp
        self.assertEqual(replaced_choice.to_json(), expected_choice.to_json())

    def test_simple_pipeline(self):
        pipeline_simple = PCA >> SelectKBest >> LogisticRegression
        simple_imputer = SimpleImputer
        replaced_pipeline = pipeline_simple.replace(PCA, simple_imputer)
        expected_pipeline = SimpleImputer >> SelectKBest >> LogisticRegression
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

    def test_choice_pipeline(self):
        pipeline_choice = (PCA | NoOp) >> SelectKBest >> LogisticRegression
        simple_imputer = SimpleImputer
        replaced_pipeline = pipeline_choice.replace(PCA, simple_imputer)
        expected_pipeline = (SimpleImputer | NoOp) >> SelectKBest >> LogisticRegression
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

    def test_planned_trained_ops(self):
        pca1 = PCA(n_components=10)
        pca2 = PCA(n_components=5)
        choice = pca1 | pca2
        pipeline_choice = (pca1 | pca2) >> LogisticRegression

        replaced_choice = choice.replace(pca1, SimpleImputer)  # SimpleImputer | pca2
        expected_choice = SimpleImputer | pca2
        self.assertEqual(replaced_choice.to_json(), expected_choice.to_json())

        replaced_pipeline = pipeline_choice.replace(
            pca1, SimpleImputer
        )  # SimpleImputer | pca2
        expected_pipeline = (SimpleImputer | pca2) >> LogisticRegression
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

        replaced_choice = choice.replace(pca2, SimpleImputer)  # pca1 | SimpleImputer
        expected_choice = pca1 | SimpleImputer
        self.assertEqual(replaced_choice.to_json(), expected_choice.to_json())

        replaced_pipeline = pipeline_choice.replace(
            pca2, SimpleImputer
        )  # pca1 | SimpleImputer
        expected_pipeline = (pca1 | SimpleImputer) >> LogisticRegression
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

        replaced_choice = choice.replace(
            PCA, SimpleImputer
        )  # SimpleImputer | SimpleImputer
        expected_choice = SimpleImputer | SimpleImputer
        self.assertEqual(replaced_choice.to_json(), expected_choice.to_json())

        replaced_pipeline = pipeline_choice.replace(
            PCA, SimpleImputer
        )  # SimpleImputer | SimpleImputer
        expected_pipeline = (SimpleImputer | SimpleImputer) >> LogisticRegression
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

    def test_nested_choice(self):
        pca1 = PCA(n_components=10)
        pca2 = PCA(n_components=5)
        pipeline_nested_choice = pca1 >> (pca1 | pca2)

        replaced_pipeline = pipeline_nested_choice.replace(pca1, SimpleImputer)
        expected_pipeline = SimpleImputer >> (SimpleImputer | pca2)
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

        replaced_pipeline = pipeline_nested_choice.replace(PCA, SimpleImputer)
        expected_pipeline = SimpleImputer >> (SimpleImputer | SimpleImputer)
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

    def test_nested_pipeline(self):
        pca1 = PCA(n_components=10)
        pca2 = PCA(n_components=5)
        first_pipeline = pca1 >> LogisticRegression
        nested_pipeline = pca1 >> (pca2 | first_pipeline)

        replaced_pipeline = nested_pipeline.replace(pca1, SimpleImputer)
        expected_pipeline = SimpleImputer >> (
            pca2 | (SimpleImputer >> LogisticRegression)
        )
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

        replaced_pipeline = nested_pipeline.replace(PCA, SimpleImputer)
        expected_pipeline = SimpleImputer >> (
            SimpleImputer | (SimpleImputer >> LogisticRegression)
        )
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

    def test_hyperparam_estimator(self):
        lr = LogisticRegression()
        linear_reg = LinearRegression()
        ada = AdaBoostRegressor(base_estimator=lr)

        replaced_ada = ada.replace(lr, linear_reg)
        expected_ada = AdaBoostRegressor(base_estimator=linear_reg)
        self.assertEqual(replaced_ada.to_json(), expected_ada.to_json())

        replaced_ada = ada.replace(LogisticRegression, linear_reg)
        expected_ada = AdaBoostRegressor(base_estimator=linear_reg)
        self.assertEqual(replaced_ada.to_json(), expected_ada.to_json())

        ada_pipeline = PCA >> SimpleImputer >> ada
        replaced_pipeline = ada_pipeline.replace(lr, linear_reg)
        expected_pipeline = (
            PCA >> SimpleImputer >> AdaBoostRegressor(base_estimator=linear_reg)
        )
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

        ada_choice = PCA | ada
        replaced_choice = ada_choice.replace(lr, linear_reg)
        expected_choice = PCA | AdaBoostRegressor(base_estimator=linear_reg)
        self.assertEqual(replaced_choice.to_json(), expected_choice.to_json())

        rfe = RFE(estimator=lr)
        replaced_rfe = rfe.replace(lr, linear_reg)
        expected_rfe = RFE(estimator=linear_reg)
        self.assertEqual(replaced_rfe.to_json(), expected_rfe.to_json())

    def test_hyperparam_estimator_list(self):
        lr = LogisticRegression()
        linear_reg = LinearRegression()
        dtc = DecisionTreeClassifier()

        cls_list = [("lr", lr), ("linear_reg", linear_reg)]
        vc = VotingClassifier(estimators=cls_list)

        replaced_vc = vc.replace(linear_reg, dtc)
        new_cls_list = [("lr", lr), ("linear_reg", dtc)]
        expected_vc = VotingClassifier(estimators=new_cls_list)
        self.assertEqual(replaced_vc.to_json(), expected_vc.to_json())

        sc = StackingClassifier(estimators=cls_list, final_estimator=vc)
        replaced_sc = sc.replace(linear_reg, dtc)
        new_cls_list = [("lr", lr), ("linear_reg", dtc)]
        expected_sc = StackingClassifier(
            estimators=new_cls_list, final_estimator=expected_vc
        )
        self.assertEqual(replaced_sc.to_json(), expected_sc.to_json())

    def test_replace_choice(self):
        choice = PCA | SelectKBest
        choice_pipeline = choice >> LogisticRegression

        replaced_pipeline = choice_pipeline.replace(choice, SelectKBest)
        expected_pipeline = SelectKBest >> LogisticRegression
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())

        choice2 = NoOp | LinearRegression
        replaced_pipeline = choice_pipeline.replace(LogisticRegression, choice2)
        expected_pipeline = choice >> choice2
        self.assertEqual(replaced_pipeline.to_json(), expected_pipeline.to_json())



import sklearn
from numpy import inf, nan
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _LinearDiscriminantAnalysisImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for LinearDiscriminantAnalysis    Linear Discriminant Analysis",
    "allOf": [
        {
            "type": "object",
            "required": [
                "solver",
                "shrinkage",
                "priors",
                "n_components",
                "store_covariance",
                "tol",
            ],
            "relevantToOptimizer": ["solver", "n_components", "tol", "shrinkage"],
            "additionalProperties": False,
            "properties": {
                "solver": {
                    "enum": ["eigen", "lsqr", "svd"],
                    "default": "svd",
                    "description": "Solver to use, possible values:   - 'svd': Singular value decomposition (default)",
                },
                "shrinkage": {
                    "anyOf": [
                        {"enum": ["auto"]},
                        {
                            "type": "number",
                            "minimumForOptimizer": 0,
                            "maximumForOptimizer": 1,
                            "minimum": 0,
                            "maximum": 1,
                            "exclusiveMinimum": True,
                            "exclusiveMaximum": True,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Shrinkage parameter, possible values:   - None: no shrinkage (default)",
                },
                "priors": {
                    "XXX TODO XXX": "array, optional, shape (n_classes,)",
                    "description": "Class priors.",
                    "enum": [None],
                    "default": None,
                },
                "n_components": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimun": 1,
                            "laleMaximum": "X/items/maxItems",
                            "minimumForOptimizer": 2,
                            "maximumForOptimizer": 256,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Number of components (< n_classes - 1) for dimensionality reduction.",
                },
                "store_covariance": {
                    "type": "boolean",
                    "default": False,
                    "description": "Additionally compute class covariance matrix (default False), used only in 'svd' solver",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Threshold used for rank estimation in SVD solver",
                },
            },
        },
        {
            "description": "shrinkage, only with 'lsqr' and 'eigen' solvers",
            "anyOf": [
                {"type": "object", "properties": {"shrinkage": {"enum": [None]}}},
                {
                    "type": "object",
                    "properties": {"solver": {"enum": ["lsqr", "eigen"]}},
                },
            ],
        },
        {
            "description": "store_covariance, only in 'svd' solver",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"store_covariance": {"enum": [False]}},
                },
                {"type": "object", "properties": {"solver": {"enum": ["svd"]}}},
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit LinearDiscriminantAnalysis model according to the given",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values.",
        },
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Project data to maximize class separation.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Input data.",
        }
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Transformed data.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict class labels for samples in X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predicted class label per sample.",
    "type": "array",
    "items": {"type": "number"},
}
_input_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Estimate probability.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Input data.",
        }
    },
}
_output_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Estimated probabilities.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}
_input_decision_function_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict confidence scores for samples.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_decision_function_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Confidence scores per (sample, class) combination",
    "laleType": "Any",
    "XXX TODO XXX": "array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.discriminant_analysis.LinearDiscriminantAnalysis#sklearn-discriminant_analysis-lineardiscriminantanalysis",
    "import_from": "sklearn.discriminant_analysis",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer", "estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
        "input_decision_function": _input_decision_function_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}
LinearDiscriminantAnalysis = make_operator(
    _LinearDiscriminantAnalysisImpl, _combined_schemas
)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.discriminant_analysis.LinearDiscriminantAnalysis#sklearn-discriminant_analysis-lineardiscriminantanalysis
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.discriminant_analysis.LinearDiscriminantAnalysis#sklearn-discriminant_analysis-lineardiscriminantanalysis
    LinearDiscriminantAnalysis = LinearDiscriminantAnalysis.customize_schema(
        covariance_estimator={
            "anyOf": [
                {
                    "type": "string",
                    "forOptimizer": False,
                },
                {"enum": [None]},
            ],
            "default": None,
            "description": "type of (covariance estimator). Estimate the covariance matrices instead of relying on the empirical covariance estimator (with potential shrinkage)",
        },
        set_as_available=True,
    )
    LinearDiscriminantAnalysis = LinearDiscriminantAnalysis.customize_schema(
        constraint={
            "description": "covariance estimator is not supported with svd solver. Try another solver",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"solver": {"not": {"enum": ["svd"]}}},
                },
                {
                    "type": "object",
                    "properties": {"covariance_estimator": {"enum": [None]}},
                },
            ],
        },
        set_as_available=True,
    )
    LinearDiscriminantAnalysis = LinearDiscriminantAnalysis.customize_schema(
        constraint={
            "description": "covariance_estimator and shrinkage parameters are not None. Only one of the two can be set.",
            "anyOf": [
                {"type": "object", "properties": {"solver": {"enum": ["svd", "lsqr"]}}},
                {
                    "type": "object",
                    "properties": {"solver": {"not": {"enum": ["eigen"]}}},
                },
                {
                    "type": "object",
                    "properties": {"covariance_estimator": {"enum": [None]}},
                },
                {"type": "object", "properties": {"shrinkage": {"enum": [None, 0]}}},
            ],
        },
        set_as_available=True,
    )

set_docstrings(LinearDiscriminantAnalysis)
